#!/usr/bin/env python
# coding: utf-8

"""
 we need to go deeper...
☺ больше ящиков богу ящиков!

č chce se ještě někam ukladat meta infa od blackboxů 
č K tomu využijme pitonovskej dak tajping a zkusíme strčit tohle mezi
č f-modelem a normálním SampleBoxem. Snad to postačí.
č Když ne - bude třeba šecko překopat.
č (však do souboru CandyBox zatím nepůjde - k tomu je potřeba podpora Readeru.)
"""

#č Nechce se mi tahnout do projektu Pandas jako rekuajred dependensi,
#č ale ani já sam pořadně nerozumím proč 
#č asi proto, že se spíše jedná o pomocnou boční "sekondari" funkcionalitu
try:
    import pandas as pd
    #č pr🐱ozatím je mi šuma ohledně toho, zda píšu do kopii, nebo do slajsu
    #č prozatím zapís do slajsu nejspíš znamená, že původní data jíž nejsou nutná
    #č prozatím se mi nechce robiť explicitné kopii, 
    #č a zatím ponechavám sber odpadů samotnému Pandas
    
    #E disable SettingWithCopyWarning 
    #E Current WellMet policy: if writing to slice have occured, it probably means
    #E original df no more needed
    pd.set_option('mode.chained_assignment', None)
except ImportError:
    print("CandyBox: error of import Pandas. CandyBox will work in Pandas-free mode")

import numpy as np

class CandyBox:

    def __new__(cls, sample_object, df=None, **kwargs):
        """
        Jedname tvrdě - není-li vstup konzistentní, 
        tak tenhle box vůbec nevytvaříme
        """
        cb = super().__new__(cls)
        cb.sampling_plan = sample_object
        
        if df is not None:
            cb.df = df
        else:
            # obalime čísly
            for key in kwargs:
                if not hasattr(kwargs[key], '__getitem__'):
                    kwargs[key] = (kwargs[key],)
                
                
            try:
                cb.df = pd.DataFrame(kwargs)
                # chcem jednoduché numpy-like chovaní
                cb.df.reset_index(inplace=True)
            except NameError: # if there is no "pandas as pd"
                cb.kwargs = kwargs
        if cb.consistency_check():
            return cb
        else:
            raise ValueError("Sample and given values hasn't the same length. Zkrátka, do sebe nepatří")
            
        
    
        
#    def __str__(cb):
#       # if pandas
#       if 'df' in cb.__dict__:
#           return 'CandyBox: %s' % cb.df
#       # if not
#       else:
#           return 'CandyBox: %s' % cb.kwargs
        
    def __repr__(cb):
        # if pandas
        if 'df' in cb.__dict__:
            # df je na svědomi pandas
            return 'CandyBox(%s, df=%s)' % (repr(cb.sampling_plan), repr(cb.df))
        # if not
        else:
            return 'CandyBox(%s, **%s)' % (repr(cb.sampling_plan), repr(cb.kwargs))
        
    def __len__(cb):
        return len(cb.sampling_plan)
        
        
    def __call__(cb, *args, **kwargs):
        # Houston, we've got a problem...
        # call meaning is different for underlaying f_models and upper Boxes
        # but SampleBox expect f_model under
        
        # hovnokód, překopat
        f = cb.sampling_plan(*args, **kwargs)
        if len(f) == 0:
            return CandyBox(f)
        else:
            return f
        
        
    
        
    def __getitem__(cb, slice):
        # if pandas
        if 'df' in cb.__dict__:
            df = cb.df.iloc[slice]
            #☺ нельзя так просто взять и накраить DataFrame
            # pandas zlobí (o kousek víc jak numpy)
            if not isinstance(df, pd.DataFrame):
                # pravděpodobně když se nám vrátí serie, tak bude interpretována jako sloup
                # fakt mi nic spolehlivějšího nenapadá
                df = pd.DataFrame(df).T
                
            # chcem jednoduché numpy-like chovaní
            # dělá problémy
            df.reset_index(drop=True, inplace=True)
            return CandyBox(cb.sampling_plan[slice], df=df)
        # if not
        else:
            sliced_dict = dict()
            for key in cb.kwargs:
                ##č nechám na uživateli, co vloží do slovníku a jak to bude slajsiť
                #č sice furt podporuju variantu s těmi slovníky
                #č ale potrebuju aby chování bylo trochu víc předvidatelné
                sliced_dict[key] = np.atleast_1d(cb.kwargs[key][slice])
            return CandyBox(cb.sampling_plan[slice], **sliced_dict)
        
    def __setattr__(cb, key, value):
        # to je vše co má samotný CandyBox
        # žádný další majetek u něho nepozoruji
        if key in ('sampling_plan', 'df', 'kwargs'):
            cb.__dict__[key] = value
        elif len(cb) == len(value):
            # if pandas
            if 'df' in cb.__dict__:
                # df je na svědomi pandas
                cb.df.loc[:, key] = value
            # if not
            else:
                cb.kwargs[key] = value
        else:
            raise ValueError("Sample and given values hasn't the same length. Zkrátka, do sebe nepatří")
        
        
    def __getattr__(cb, attr):
        # branime sa rekurzii
        # defend against recursion
        if attr == 'sampling_plan':
            raise AttributeError
            
        elif attr == 'candybox':
            return cb
            
        # hledáme obraceně
        # nejdřív se zeptame f_model,
        # teprve když ten nič nemá  
        # mrkneme atribut u sebe
        try:
            return getattr(cb.sampling_plan, attr)
        except AttributeError:
            return cb._lookup(attr)
            
    def _lookup(cb, attr):
        # if pandas
        if 'df' in cb.__dict__:
            if attr in cb.df:
                # zkusím převést na numpy, protože 
                # Pandasovy indexy jen zlobí
                # a stejně nikdo je nepouživá a na ně nespolehá
                # další kód by neměl furt řešit, co ta bombonjera má uvnitř
                # i když je to škoda, že nepujde měnit jednotlivé hodnoty
                try:
                    return cb.df[attr].to_numpy()
                except AttributeError:
                    return np.array(cb.df[attr])
                
                # zkusím nepřevadět na numpy, protože chcu mäť možnost hodnoty měnit
                # pri slajsech resetuju indexy, takže musí to být v pohodě
                # doufám, že tím нищего не поломаju
                # šecko se tím zkazilo a je taky otázkou 
                # zda je to vhodně když se furt robej slajsy
                # nechám to na uživatelském kódu
                #return cb.df[attr]
            else: # implicitně pandas hodí KeyError, kterej nechcem
                raise AttributeError
        
        # if not
        elif attr in cb.kwargs:
            value = cb.kwargs[attr]
            if len(cb)==len(value):
                return cb.kwargs[attr]
            else:
                # у нас есть дата, но мы их вам недадим)
                # zde nechť bude KeyError
                raise KeyError("CandyBox: well, we have some data, but they are not consistent, so we haven't")
                
        else: # implicitně slovníky (stejně jako pandas) hazej KeyError, kterej nechcem
            raise AttributeError
            
        
        
    def add_sample(cb, input):
        """
        hlavní požadavek - jsou-li samply uspěšně sjednoceny,
        tak ty blbé zbytky nesmejí mi hodit chybu!
        котлеты - отдельно, мухи - отдельно
        """
        
        # čo to je za vstup?
        sweety_input = hasattr(input, 'sampling_plan')
            
        #
        # котлеты
        #
        # nechcu zde try-catch blok
        if sweety_input:
            cb.sampling_plan.add_sample(input.sampling_plan)
        else:
            cb.sampling_plan.add_sample(input)
        
        
        # мухи
        if sweety_input:
            # if pandas
            if 'df' in cb.__dict__:
                #if 'df' in input.__dict__: # jen pro formu kontrola
                # pandas musí mít i tamtenhle objekt, žejo?
                cb.df = cb.df.append(input.df, ignore_index=True)
            # if not
            else:
                # zjednodušený join
                sample_len = len(cb.sampling_plan)
                for key in cb.kwargs:
                    # ani nebudu kontrolovat
                    #if isinstance(key, np.ndarray):
                        
                    if key in input.kwargs:
                        cb.kwargs[key] = np.append(cb.kwargs[key], input.kwargs[key])
                    else:
                        fill_len = sample_len - len(cb.kwargs[key])
                        full = (None for __ in range(fill_len))
                        cb.kwargs[key] = np.append(cb.kwargs[key], (*full,))
                
                
        else: # nesladký vstup
            sample_len = len(cb.sampling_plan)
            # if pandas
            if 'df' in cb.__dict__:
                fill_len = sample_len - len(cb.df)
                full_df = pd.DataFrame(index=range(fill_len))
                cb.df = cb.df.append(full_df, ignore_index=True)
            # if not
            else:
                for key in cb.kwargs:
                    fill_len = sample_len - len(cb.kwargs[key])
                    full = (None for __ in range(fill_len))
                    cb.kwargs[key] = np.append(cb.kwargs[key], (*full,))

                
    #E we'll see, if .new_sample will be needed
    #E year, we need it
    #♥ мыным выль сэмпл кулэ!
    def new_sample(cb, input=None, space='R'):
        # už je to stavá matoucím
        # pokud input je, děláme tohle
        # vrácíme samy sobě
        if input is None:
            return CandyBox(cb.sampling_plan())
            
        # pokud input není, děláme vonohle
        # konvertujem vstup na naše rozdělení
        else:
            # čo to je za vstup?
            if hasattr(input, 'sampling_plan'): #sweety_input:
                # if pandas
                if hasattr(input, 'df'):
                    return CandyBox(cb.sampling_plan.new_sample(input.sampling_plan, space=space), df=input.df)
                    
                # if there is no pandas DataFrame
                # we suppose there is dictionary in input.kwargs
                else:
                    return CandyBox(cb.sampling_plan.new_sample(input.sampling_plan, space=space), **input.kwargs)
                
            # nesladký vstup
            else:
                return cb.sampling_plan.new_sample(input, space)
        
        
        
            
            
        
    def consistency_check(cb):
        # řvat na celé město nebudeme
        # if pandas
        if 'df' in cb.__dict__:
            return len(cb.sampling_plan)==len(cb.df)
        # if not
        else:
            sample_len = len(cb.sampling_plan)
            return all(sample_len == len(cb.kwargs[key]) for key in cb.kwargs)



#!/usr/bin/env python
# coding: utf-8

import gc
import pyqtgraph as pg
from pyqtgraph.Qt import QtGui, QtWidgets
from pyqtgraph.Qt import QtCore

import numpy as np
from scipy import spatial # for distance matrix
from .. import wireframe
from .. import reader


#č pg trik, aby GC nám nevymazal paměť
windows = set()

### Define a top-level widget to hold everything
class MatrixWindow(QtWidgets.QMainWindow):
    #č box_runned dublikuje slice_changed
    # do not redraw twice!
    box_runned = QtCore.pyqtSignal()
    space_changed = QtCore.pyqtSignal(str)
    metric_changed = QtCore.pyqtSignal(str)
    slice_changed = QtCore.pyqtSignal(int)
    redraw_called = QtCore.pyqtSignal()
    
    
    def __init__(self, sample_box, space='R', parent=None, *args, **kwargs):
        try:
            self.box_runned = parent.box_runned
            self.app = parent.app
        except:
            self.app = pg.mkQApp() #оӵ кулэ-а?
            ## Start the Qt event loop
            self.app.exec_()
            
        super().__init__()
        
        #č pg trik, aby GC nám nevymazal paměť
        windows.add(self)
        
        
        
        try:
            self.setWindowTitle("%sD: %s" %(sample_box.nvar, sample_box.gm_signature))
        except:
            self.setWindowTitle("%sD nodes" % sample_box.nvar)
        
        # for debug
        # container for errors
        # to trace errors
        self.errors = []
        
        self.kwargs = kwargs
        self.sample_box = sample_box
        self.space = space
        self.metric = 'euclidean'
        
        # "zapnuté" prostory
        #self.spaces = ['R', 'aR', 'Rn', 'aRn', 'P', 'aP', 'GK', 'aGK', 'G', 'aG', 'U', 'aU']
        self.spaces = ['R', 'aR', 'Rn', 'aRn', 'P', 'GK', 'G', 'aG', 'U', 'aU']
        
        #self.metrics = ['1', '2', 'inf']
        self.metrics = ['euclidean', 'cityblock', 'chebyshev',
                        'seuclidean', 'sqeuclidean', 'cosine',
                        'correlation', 
                        'canberra', 'braycurtis', 'mahalanobis' ]
        
        # common setup
        self.setup()
        
        ## Display the widget as a new window
        self.show()

        #
        self.redraw_called.emit()
        
        
    
    def closeEvent(self, event):
        #č nejsem 100% jist co dělám.
        #č Ale když už ImageView API 
        #č vysloveně nabízí "korektní" .close()
        #č tak deme do toho
        self.image_view.close()
        windows.discard(self)
        event.accept() # let the window close
    
    
    def setup_menubar(self):    
        self.bar = self.menuBar()
        self.file_menu = self.bar.addMenu("File")
        #č kontejner jen aby Python mně Cečkové objekty nevymazal
        self.file_actions = [] 
        load_coordinates_action = QtGui.QAction("Load coordinates", self)
        load_coordinates_action.triggered.connect(self.load_coordinates)
        self.file_menu.addAction(load_coordinates_action)
        self.file_actions.append(load_coordinates_action)
        
        export_coordinates_action = QtGui.QAction("Export coordinates", self)
        export_coordinates_action.triggered.connect(self.export_coordinates)
        self.file_menu.addAction(export_coordinates_action)
        self.file_actions.append(export_coordinates_action)
        
        import_samples_action = QtGui.QAction("Import samples", self)
        import_samples_action.triggered.connect(self.import_samples)
        self.file_menu.addAction(import_samples_action)
        self.file_actions.append(import_samples_action)
        
        export_samples_action = QtGui.QAction("Export samples", self)
        export_samples_action.triggered.connect(self.export_samples)
        self.file_menu.addAction(export_samples_action)
        self.file_actions.append(export_samples_action)
        
        self.view = self.bar.addMenu("View")
        
        self.graph_menu = self.bar.addMenu("Box")
        self.add_random_points_action = QtGui.QAction("Add random points", self)
        self.add_random_points_action.triggered.connect(self.add_random_points)
        self.graph_menu.addAction(self.add_random_points_action)
        self.batch_run_action = QtGui.QAction("Batch run", self)
        self.batch_run_action.triggered.connect(self.batch_run)
        self.graph_menu.addAction(self.batch_run_action)
        
        
    def get_sample_box(self):
        nsim = self.slider.value()
        if nsim == self.sample_box.nsim:
            return self.sample_box
        else:
            return self.sample_box[:nsim]
    
    
    # intended as a common setup function
    def setup(self):    
        self.setup_menubar()
        
        ### Create some widgets to be placed inside
        self.combo_metric = pg.ComboBox(items=self.metrics, default=self.metric)
        self.combo_metric.activated[str].connect(self.change_metric)
        
        self.combo_space = pg.ComboBox(items=self.spaces, default=self.space)
        self.combo_space.activated[str].connect(self.change_space)
        
        
        self.label_nsim = QtWidgets.QLabel()
        self.label_nsim.setText(str(self.sample_box.nsim))
        
        self.slider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        self.slider.valueChanged.connect(self._slider_chainged)
        self.slider.sliderReleased.connect(lambda:self.slice_changed.emit(self.slider.value()))
        self.slider.setMaximum(self.sample_box.nsim)
        self.slider.setValue(self.sample_box.nsim)
        
        #č jen aby se slider probral, když uživatel ručně přídá bodíky
        self.redraw_called.connect(self.wake_up_slider)
        self.box_runned.connect(self.wake_up_slider)
        
        self.btn = QtWidgets.QPushButton('run box')
        self.btn.clicked.connect(self.run_sb)
        
#        self.btn2 = QtWidgets.QPushButton('connect/disconnect')
#        self.btn2.setCheckable(True)
#        self.btn2.clicked.connect(self.bx_connect)

        ## Create a grid layout to manage the widgets size and position
        self.layout = pg.LayoutWidget()
        self.setCentralWidget(self.layout)
        
        self.list_view = QtWidgets.QListWidget()

        self.image_view = MView(self, parent=self.layout)
        
        self.btn3 = QtWidgets.QPushButton('auto levels')
        self.btn3.clicked.connect(lambda:self.image_view.autoLevels())
        
        

        
        
        
        ## Add widgets to the layout in their proper positions
        #self.layout.addWidget(self.list_view, 0, 0, 2, 1) 
        self.layout.addWidget(self.combo_metric, 0, 0)  
        self.layout.addWidget(self.combo_space, 0, 1)   
        self.layout.addWidget(self.slider, 0, 2)   
        self.layout.addWidget(self.label_nsim, 0, 3)
        self.layout.addWidget(self.btn, 0, 4) 
        #self.layout.addWidget(self.btn2, 0, 4)
        self.layout.addWidget(self.btn3, 0, 5) 
        self.layout.addWidget(self.image_view, 1, 0, 1, 6)
        


        # status bar, mainly to observe BlackBox
        self.statusBar = QtWidgets.QStatusBar()
        self.setStatusBar(self.statusBar)
        
        self.dm = DistanceMatrix(self)
        
        # Dockables, najzajimavejší věc
        self.dockables = []
        
        
        
        dock = QtWidgets.QDockWidget("View", self)
        dock.setWidget(self.list_view)
        self.dockables.append(dock)
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea, dock)
        
        
        dock = QtWidgets.QDockWidget("Contacts", self)
        dock.setWidget(ContactWidget(self, dock))
        self.dockables.append(dock)
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea, dock)
        
        
        
        for dock in self.dockables:
            self.view.addAction(dock.toggleViewAction())
            #dock.setFloating(True)
    
    
    csv_filter = "CSV files (*.csv)"
    npy_filter = "NumPy binary files (*.npy)"
    txt_filter = "Text files (*.txt)"
    gz_filter = "Compressed text files (*.gz)"
       
    def load_coordinates(self):
        """
        Loads numpy data and adds them to the sample_box
        """ 
        nodes = None
        filter = self.npy_filter + ';;' + self.txt_filter + ';;' + self.gz_filter
        wt = self.sample_box
        
        try:
            filename, filter = pg.FileDialog.getOpenFileName(
                    parent=self, 
                    caption="Load coordinates",
                    directory="store/%s_%sD.npy" % (wt.gm_signature, wt.nvar),
                    filter=filter, 
                    initialFilter=self.npy_filter,
                    #options=pg.FileDialog.Option.DontConfirmOverwrite
                    )
            if filename:
                if filter == self.npy_filter:
                    raw_nodes = np.load(filename) 
                else:
                    raw_nodes = np.loadtxt(filename) 
                
                # create sample
                nodes = wt.f_model.new_sample(sample=raw_nodes, \
                                    space=self.space, extend=False)
                
        except BaseException as e:
            print(self.__class__.__name__ + ":", \
                            " coordinates loading failed", repr(e))
                            
        #č zde jíž nemá cenu chytat chyby. Sbohem!
        if nodes is not None:
            # run!
            wt.run_sample(nodes)
            self.box_runned.emit()
        
        
    def export_coordinates(self):
        """
        Exports current sample box coordinates
        """ 
        filter = self.npy_filter + ';;' + self.txt_filter + ';;' + self.gz_filter
        wt = self.sample_box
        
        filename, filter = pg.FileDialog.getSaveFileName(
                parent=self, 
                caption="Export coordinates",
                directory="store/%s_%sD.npy" % (wt.gm_signature, wt.nvar),
                #filter=csv_filter,
                filter=filter, 
                initialFilter=self.npy_filter,
                #options=pg.FileDialog.Option.DontConfirmOverwrite
                )
        if filename:
            nsim = self.slider.value()
            nodes = self.sample_box.f_model[:nsim]
            raw_nodes = getattr(nodes, self.space)
            try:
                if filter == self.npy_filter:
                    np.save(filename, raw_nodes, allow_pickle=False)
                else:
                    np.savetxt(filename, raw_nodes) 
            except BaseException as e:
                print(self.__class__.__name__ + ":", \
                            " coordinates export failed", repr(e)) 
        
        
        
    def import_samples(self):
        wt = self.sample_box
        
        filename, __filter = pg.FileDialog.getOpenFileName(
                parent=self, 
                caption="Import samples",
                directory="store/%s_%sD.csv" % (wt.gm_signature, wt.nvar),
                #filter=csv_filter,
                filter=self.csv_filter, # + ";;All files(*)", 
                initialFilter=self.csv_filter,
                #options=pg.FileDialog.Option.DontConfirmOverwrite
                )
        if filename and (filename[-4:]=='.csv'):
            #č Reader vždy dopíše '.csv' na konci souboru
            #č nechcu to teď měnit
            filename = filename[:-4]
            
            #č nechcu chytat žádné chyby. Sbohem!
            sample_box = reader.reader(filename, f_model=wt.f_model)
            wt.add_sample(sample_box)
            self.box_runned.emit()

    def export_samples(self):
        wt = self.sample_box
        
        filename, __filter = pg.FileDialog.getSaveFileName(
                parent=self, 
                caption="Export samples",
                directory="store/%s_%sD.csv" % (wt.gm_signature, wt.nvar),
                #filter=csv_filter,
                filter=self.csv_filter, # + ";;All files(*)", 
                initialFilter=self.csv_filter,
                #options=pg.FileDialog.Option.DontConfirmOverwrite
                )
        if filename:
            #č Reader vždy dopíše '.csv' na konci souboru
            #č nechcu to teď měnit
            if filename[-4:]=='.csv':
                filename = filename[:-4]
            nsim = self.slider.value()
            reader.export(filename, wt[:nsim])


    #
    #č Tlačítka!
    #

    # INHERITED by gl_plot
    def _slider_chainged(self, value):
        #č .setMaximum() nezpůsobuje emitování slice_changed, jsem zkontroloval
        self.slider.setMaximum(self.sample_box.nsim)
        self.label_nsim.setText(str(value))
        if not self.slider.isSliderDown(): # co to vůbec děla?
            self.slice_changed.emit(value)
            
    
    def change_space(self, space):
        self.space = space
        self.space_changed.emit(space)
        
    def change_metric(self, metric):
        self.metric = metric
        self.metric_changed.emit(metric)
        
    # INHERITED by gl_plot
    def run_sb(self):
        with pg.BusyCursor():
            self.last_shot = self.sample_box()
            self.box_runned.emit()
    
    
    def wake_up_slider(self):
        # slider
        #č zpusobí slice_changed
        self.slider.setMaximum(self.sample_box.nsim)
        self.slider.setValue(self.sample_box.nsim)
        
    def add_random_points(self):
        ns, ok = QtWidgets.QInputDialog.getInt(
                    self,"Add random points","number", value=1, min=1)
        if ok:
            wt = self.sample_box
            nodes = wt.f_model(ns)
            wt.run_sample(nodes)
            self.box_runned.emit()
        
    def batch_run(self):
        runs, ok = QtWidgets.QInputDialog.getInt(
                    self,"Batch run","runs", value=100, min=1)
  
        if ok:
            with pg.ProgressDialog("Running..", 0, runs,\
                         cancelText='Stop', busyCursor=True) as dlg:
                for i in range(runs):
                    # keep the GUI responsive :)
                    self.app.processEvents()
                    
                    self.last_shot = self.sample_box()
                
                    # slider
                    #č zpusobí slice_changed
                    self.slider.setMaximum(self.sample_box.nsim)
                    self.slider.setValue(self.sample_box.nsim)
                    
                    self.box_runned.emit()
                    
                    dlg += 1
                    if dlg.wasCanceled():
                        break
                    

        

# piece of code from pyqtgraph examples
class CustomViewBox(pg.ViewBox):
    ## reimplement mouseDragEvent to disable continuous axis zoom
    def mouseDragEvent(self, ev, axis=None):
        if ev.button() == QtCore.Qt.MouseButton.RightButton:
            ev.ignore()
        else:
            pg.ViewBox.mouseDragEvent(self, ev, axis=axis)




right_button = QtCore.Qt.MouseButton.RightButton
left_button = QtCore.Qt.MouseButton.LeftButton

class MView(pg.ImageView):
    mouse_moved = QtCore.pyqtSignal(int, int)
    mouse_clicked = QtCore.pyqtSignal(int, int)
    mouse_double_clicked = QtCore.pyqtSignal(int, int)
    mouse_right_dragged = QtCore.pyqtSignal(int, int)
    
        
    def __init__(self, w, *args, **kwargs):
        self.vb = CustomViewBox()
        self.pi = pg.PlotItem(viewBox=self.vb)
        super().__init__(*args, view=self.pi, **kwargs) #axisOrder='row-major',
        
        w.slice_changed.connect(self.on_slice_changed)
        self.app = w.app
        
        self.nsim = nsim = w.slider.value()
        
        self.data = np.empty((3, nsim, nsim))
        self.data[2] = -1 #č kvuli ContactWidgetu
        
        self.show()
        #self.setImage(data, levelMode='rgba')
        #self.setBackground('w')
        
        
        self.ar_item = QtWidgets.QListWidgetItem('auto range')
        self.ar_item.setFlags(self.ar_item.flags() | QtCore.Qt.ItemIsUserCheckable)
        self.ar_item.setCheckState(QtCore.Qt.Checked)
        w.list_view.addItem(self.ar_item)
        
        self.al_item = QtWidgets.QListWidgetItem('auto levels')
        self.al_item.setFlags(self.al_item.flags() | QtCore.Qt.ItemIsUserCheckable)
        self.al_item.setCheckState(QtCore.Qt.Checked)
        w.list_view.addItem(self.al_item)
        
        
        
        self.proxy  = pg.SignalProxy(self.pi.scene().sigMouseMoved, 
                                    rateLimit=60, slot=self.on_mouse_moved)
        self.x = -1
        self.y = -1
                                    
        self.pi.scene().sigMouseClicked.connect(self.on_click)
        
    
    def __setattr__(self, attr, value):
        if attr == 'red':
            self.data[0] = value
        elif attr == 'green':
            self.data[1] = value
        elif attr == 'blue':
            self.data[2] = value
        else:
            self.__dict__[attr] = value
            
            
            
    def __getattr__(self, attr):
        if attr == 'red':
            return self.data[0]
        if attr == 'green':
            return self.data[1]
        if attr == 'blue':
            return self.data[2]
        raise AttributeError(attr)
    
    
    def on_slice_changed(self, nsim):
        #č resize nejde. Ani s gc. Asi pyqtgraph drží referenci
        #č i když já pekně vím, že ta data kopíruje a pro Qt normalizuje
        self.data = np.empty((3, nsim, nsim))
        #č zelené-červené barvy přepíše distance matrix
        #č za modrou ale nikdo zodpovednost nenese
        self.data[2] = -1 #č kvuli ContactWidgetu 
        self.nsim = nsim
    
    
    def update(self, **kwargs):
        if 'autoRange' not in kwargs:
            kwargs['autoRange'] = self.ar_item.checkState()
        if 'autoLevels' not in kwargs:
            kwargs['autoLevels'] = self.al_item.checkState()
        
        self.setImage(self.data.T, levelMode='rgba', **kwargs)
        
    
    def on_mouse_moved(self, evt):
        pos = evt[0]  ## using signal proxy turns original arguments into a tuple
        if self.pi.sceneBoundingRect().contains(pos):
            mousePoint = self.pi.vb.mapSceneToView(pos)
            x = int(mousePoint.x())
            y = int(mousePoint.y())
            if (x >= 0) and (x < self.nsim) and (y >= 0) and (y < self.nsim):
                self._mouse_moved(x, y)
            else:
                self._mouse_moved(-1, -1)
        else:
            self._mouse_moved(-1, -1)
            
    def _mouse_moved(self, x, y):
        if (x != self.x) or (y != self.y):
            self.x = x
            self.y = y
            self.mouse_moved.emit(x, y)
            if self.app.mouseButtons() == right_button:
                self.mouse_right_dragged.emit(x, y)
            
            
    
    def on_click(self, evt):
        if self.x == -1:
            return None
        if evt.double():
            self.mouse_double_clicked.emit(self.x, self.y)
        elif evt.button() == left_button:
            self.mouse_clicked.emit(self.x, self.y)
            



class DistanceMatrix:
    def __init__(self, w):
        self.w = w
        
        w.image_view.mouse_moved.connect(self.on_mouse_moved)
        
        w.metric_changed.connect(self.on_some_stuff_changed)
        w.space_changed.connect(self.on_some_stuff_changed)
        w.slice_changed.connect(self.on_nsim_changed)
        w.redraw_called.connect(self.on_some_stuff_changed)
        
        n = w.slider.value() 
        self.condensed_matrix = np.empty(n * (n - 1) // 2)
        self.nsim = n
        
    def on_nsim_changed(self, n):
        self.condensed_matrix.resize(n * (n - 1) // 2)
        self.nsim = n
        
        self.on_some_stuff_changed()
    
    def on_some_stuff_changed(self, *args, **kwargs):
        sample_box = self.w.get_sample_box()
        nsim = sample_box.nsim
        if nsim > 0:
            X = getattr(sample_box, self.w.space)
            spatial.distance.pdist(X, self.w.metric, out=self.condensed_matrix)
            
            mask = np.zeros((nsim, nsim))
            
            failure_points = sample_box.failure_points 
            mask[failure_points] += 0.5
            mask[:, failure_points] += 0.5
            
            dmatrix = spatial.distance.squareform(self.condensed_matrix)
            
            self.w.image_view.red = dmatrix * mask
            self.w.image_view.green = dmatrix * (1 - mask)
            self.w.image_view.update()
        
        self.w.statusBar.showMessage("")
    
    
    def on_mouse_moved(self, x, y):
        if x == -1:
            self.w.statusBar.showMessage("")
            return None
            
        if x != y:
            m = self.nsim
            i, j = min(x, y), max(x, y)
            entry = m * i + j - ((i + 2) * (i + 1)) // 2
            val = self.condensed_matrix[entry]
        else:
            val = 0
        msg = "distance between %d and %d is %g" % (x, y, val)
        self.w.statusBar.showMessage(msg)







class ContactWidget(pg.LayoutWidget):
    def __init__(self, w,  parent=None):
        super().__init__(parent)
        self.w = w
        
        w.image_view.mouse_moved.connect(self.on_mouse_moved)
        w.image_view.mouse_clicked.connect(self.on_mouse_dragged)
        w.image_view.mouse_double_clicked.connect(self.on_double_click)
        w.image_view.mouse_right_dragged.connect(self.on_mouse_dragged)
        
        w.space_changed.connect(self.on_space_changed)
        w.slice_changed.connect(self.on_nsim_changed)
        #w.redraw_called.connect(self.hide)
        
        n = w.slider.value() 
        size = n * (n - 1) // 2
        self.condensed_contacts = np.zeros(size, dtype=np.int8)
        self.mask = np.tri(n, k=-1, dtype=bool).T
        self.nsim = n
        
        self.setup()
        self.kwargs = {}
        self.check_contact = self.peacefully_check_contact
        self.got_user_consent = False
        self.setup_CS()
        
    def on_nsim_changed(self, n):
        try:
            del self.qframe
            del self.condensed_qontacts
        except AttributeError:
            pass
        size = n * (n - 1) // 2
        gc.collect()
        self.condensed_contacts.resize(size)
        self.condensed_contacts[:] = 0
        self.mask = np.tri(n, k=-1, dtype=bool).T
        self.nsim = n
        
        self.on_space_changed()
    
    def on_space_changed(self, *args, **kwargs):
        self.setup_CS()
    
    
    def on_mouse_moved(self, x, y):
        if (x == -1) or (x == y):
            self.status_label.setText("")
            return None
            
        m = self.nsim
        i, j = min(x, y), max(x, y)
        entry = m * i + j - ((i + 2) * (i + 1)) // 2
        try:
            if self.condensed_qontacts[entry]:
                self.status_label.setText("There is DEFINITELY contact between %d and %d" % (x, y))
            else:
                self.status_label.setText("There is definitely NO contact between %d and %d" % (x, y))
        except AttributeError:
            val = self.condensed_contacts[entry]
            if val > 0:
                self.status_label.setText("There IS contact between %d and %d" % (x, y))
            elif val < 0:
                self.status_label.setText("There is NO contact between %d and %d" % (x, y))
            else:
                self.status_label.setText("")
        
        
    def on_mouse_dragged(self, x, y):
        if (x == -1) or (x == y):
            return None
            
        m = self.nsim
        i, j = min(x, y), max(x, y)
        entry = m * i + j - ((i + 2) * (i + 1)) // 2
        val = self.check_contact(entry, i, j)
        self.show()
        
        if val > 0:
            self.status_label.setText("There IS contact between %d and %d" % (x, y))
        else:
            self.status_label.setText("There is NO contact between %d and %d" % (x, y))
            
    def on_double_click(self, x, y):
        if (x == -1) or (x == y):
            return None
            
        m = self.nsim
        i, j = min(x, y), max(x, y)
        entry = m * i + j - ((i + 2) * (i + 1)) // 2
        val = self.force_check_contact(entry, i, j)
        self.show()
        
        if val > 0:
            self.status_label.setText("There IS contact between %d and %d" % (x, y))
        else:
            self.status_label.setText("There is NO contact between %d and %d" % (x, y))
            
        
    def discover(self):
        nsim = self.nsim
        if  nsim > 0:
            self.stopbtn.setEnabled(True)
            self.stopbtn.setCheckable(True)
            
            if self.param.getValues()['check_events'][0] == 'All pairs':
                for i in range(nsim-1, -1, -1):
                    # keep the GUI responsive :)
                    self.show()
                    self.w.app.processEvents()
                    if self.stopbtn.isChecked():
                        break
                    preentry = nsim * i - ((i + 2) * (i + 1)) // 2
                    for j in range(i+1, nsim):
                        self.check_contact(preentry + j, i, j)
            elif self.param.getValues()['check_events'][0] == 'Failure and mixed pairs':
                failsi = self.w.get_sample_box().failsi
                for i in range(nsim-1, -1, -1):
                    # keep the GUI responsive :)
                    self.show()
                    self.w.app.processEvents()
                    if self.stopbtn.isChecked():
                        break
                    preentry = nsim * i - ((i + 2) * (i + 1)) // 2
                    if failsi[i]: #č první je červený
                        for j in range(i+1, nsim):
                            self.check_contact(preentry + j, i, j)
                    else:
                        for j in range(i+1, nsim):
                            if failsi[j]:
                                self.check_contact(preentry + j, i, j)
            else: # 'Mixed pairs only'
                failsi = self.w.get_sample_box().failsi
                for i in range(nsim-1, -1, -1):
                    # keep the GUI responsive :)
                    self.show()
                    self.w.app.processEvents()
                    if self.stopbtn.isChecked():
                        break
                    preentry = nsim * i - ((i + 2) * (i + 1)) // 2
                    if failsi[i]: #č první je červený
                        for j in range(i+1, nsim):
                            if not failsi[j]:
                                self.check_contact(preentry + j, i, j)
                    else:
                        for j in range(i+1, nsim):
                            if failsi[j]:
                                self.check_contact(preentry + j, i, j)
            self.stopbtn.setChecked(False)
            self.stopbtn.setEnabled(False)
            self.show()


    def force_check_contact(self, entry, i, j):
        val = -1 + 2 * self.CS.is_couple((i, j), **self.kwargs)
        self.condensed_contacts[entry] = val
        #č není to úplně ideální kód style. Nastavit a zaroveň vrácet :(
        return val 
        
    
    def peacefully_check_contact(self, entry, i, j):
        val = self.condensed_contacts[entry]
        if val == 0:
            return self.force_check_contact(entry, i, j)
        else:
            return val
        
        
    def setup(self):
        self.toolbar = QtWidgets.QToolBar(self)
        
        action = self.toolbar.addAction("discover", self.discover)
        btn = self.toolbar.widgetForAction(action)
        btn.setAutoRaise(False)
        btn.setToolTip("Searches for a common facets between Voronoi cells")
        
        action = self.toolbar.addAction("stop", self.stop)
        self.stopbtn = btn = self.toolbar.widgetForAction(action)
        btn.setAutoRaise(False)
        btn.setCheckable(True)
        btn.setChecked(False)
        btn.setEnabled(False)
        
        
        action = self.toolbar.addAction("Qhull check", self.check)
        btn = self.toolbar.widgetForAction(action)
        btn.setAutoRaise(False)
        btn.setToolTip("Compare with exact Qhull wireframe")
        
        action = self.toolbar.addAction("show", self.show)
        btn = self.toolbar.widgetForAction(action)
        btn.setAutoRaise(False)
        
        action = self.toolbar.addAction("hide", self.hide)
        btn = self.toolbar.widgetForAction(action)
        btn.setAutoRaise(False)
        
        action = self.toolbar.addAction("clear", self.clear)
        btn = self.toolbar.widgetForAction(action)
        btn.setAutoRaise(False)
        
        action = self.toolbar.addAction("mask", self.clear_non_blue)
        btn = self.toolbar.widgetForAction(action)
        btn.setAutoRaise(False)
        
        self.addWidget(self.toolbar, row=0, col=0)
        
        
        #оӵ остальной (люкет) уллапала
        
        ### Create ParameterTree widget
        self.ptree = pg.parametertree.ParameterTree()
        self._set_param()
        self.ptree.setParameters(self.param, showTop=False)
        
        self.addWidget(self.ptree, row=1, col=0)
        
        self.status_label = QtWidgets.QLabel()
        #self.status_label.setText()
        self.addWidget(self.status_label, row=2, col=0)
        
    def stop(self):
        self.stopbtn.setChecked(True)
        
    def qframe_callback(self):
        # keep the GUI responsive :)
        self.w.app.processEvents()
        
        self.dlg += 1
        return self.dlg.wasCanceled()
  
    
            
    @staticmethod
    def get_user_consent():
        return QtWidgets.QMessageBox.question(None, 
            "Performing triangulation in high dimensions may take a while...",
            "Would you like to start Qhull anyway?"
            ) == QtWidgets.QMessageBox.Yes
        
        
    def get_qhull(self):
        try:
            return True, self.qframe
        except AttributeError:
            sample_box = self.w.get_sample_box()
            if (sample_box.nvar < 10) or self.got_user_consent or self.get_user_consent():
                self.got_user_consent = True
                sample_space = getattr(sample_box, self.w.space)
                self.qframe = wireframe.Qframe(sample_space)
                return True, self.qframe
            else:
                return False, None
        
        
    def check(self):
        if self.get_qhull()[0]:
            with pg.ProgressDialog("Going over all the simplices..", 0, 
                                    self.qframe.nsimplex, cancelText='Stop', 
                                    busyCursor=True) as dlg:
                self.dlg = dlg
                self.qframe.generate_wireframe(self.qframe_callback)
                if not self.dlg.wasCanceled():
                    self.condensed_qontacts = self.qframe.condensed_contacts
                    self.show()
            
        
        
    def clear(self):
        self.condensed_contacts[:] = 0
        try:
            del self.qframe
            del self.condensed_qontacts
        except AttributeError:
            pass
            
        self.w.image_view.blue[self.mask] = -1
        self.w.image_view.update()
    
    def clear_non_blue(self):
        vmask = self.condensed_contacts != 1
        Mmask = spatial.distance.squareform(vmask)
        self.w.image_view.red[Mmask] = 0
        self.w.image_view.green[Mmask] = 0
        self.w.image_view.update()
    
    def show(self):
        try:
            self.w.image_view.blue[self.mask] = self.condensed_qontacts * 2 - 1
            self.w.image_view.blue[self.mask] -= self.condensed_contacts
            self.w.image_view.blue[self.mask] /= 2
        except AttributeError:
            self.w.image_view.blue[self.mask] = self.condensed_contacts
        self.w.image_view.update(autoRange=False)
    
    def hide(self):
        self.w.image_view.blue[self.mask] = -1
        self.w.image_view.update()

        
    def _set_param(self):
        params = list()
        #č check_ ovlivňuje discover(). discover ho i kontroluje
        params.append({'name': 'check_events', 'title': 'pairs to search', 
                        'type': 'list', 'value': True,
                        'values': ['All pairs', 'Failure and mixed pairs', 'Mixed pairs only'],
                        'tip': 'adjacency search for specified pairs'})
        params.append({'name': 'force_update', 'title': 'reevaluate', 
                        'type': 'bool', 'value': False,
                        'tip': "always reevaluate even discovered contacts"})
        params.append({'name': 'method', 'type': 'list', 'value': 'ConvexLinear', 
                     'values': ['DirectContact', 'Gabriel', 'ConvexSolver', 'ConvexLinear', 'LinearSolver', 'LocalizedHull', 'Qhull'],
                     'title': "method"})
        
        params.append({'name': 'solver', 'title': "solver", 'type': 'list', 'value': 'convex_sprite',
                    'values': ['convex_solver', 'convex_spline', 'convex_sort', 'convex_sprite', 'convex_slice'],  
                    'tip': "Versions from the earliest (the worst) to the latest (the best). Names have no any special meaning."})
                     
        params.append({'name': 'tries_to_fix', 'type': 'int', 
                    'limits': (0, float('inf')), 'value': 1, 'default': 1, 
                    'title': "additional tries", 
                    'tip': "Number of additional (over required ndim) tries to rotate hyperplane and find the contact."})   
            
        params.append({'name': 'tol', 'type': 'float', 
                    'title': "tolerance", 
                    'limits': (0, float('inf')), 'value': 1e-5, 'default': 1e-7})   
                    
        self.param = pg.parametertree.Parameter.create(name='params', type='group', children=params)
        
        #č chlapi tomu svému stromu dali zabrat
        #č udělali i signaly
        p = self.param.child('force_update')
        p.sigValueChanged.connect(self.check_method_changed)
        
        p = self.param.child('solver')
        p.sigValueChanged.connect(self.cs_solver_changed)
        
        p = self.param.child('method')
        p.sigValueChanged.connect(self.method_changed)
        
        
        p = self.param.child('tries_to_fix')
        p.sigValueChanged.connect(self.ttf_changed)
        
        p = self.param.child('tol')
        p.sigValueChanged.connect(self.tol_changed)
        
    
    def ttf_changed(self, parameter):
        self.kwargs['tries_to_fix'] = parameter.value()
        
    def tol_changed(self, parameter):
        self.kwargs['tol'] = parameter.value()
        
        
    def check_method_changed(self, parameter):
        if parameter.value():
            self.check_contact = self.force_check_contact
        else:
            self.check_contact = self.peacefully_check_contact
        
    def cs_solver_changed(self, parameter):
        #č ale.. lezeme do vnítřku cizí třídy..
        solver = getattr(wireframe, parameter.value())
        self.CS.convex_solver = solver
        
    def method_changed(self, parameter):
        self.setup_CS()
        method = parameter.value()
        if method in ('ConvexSolver', 'ConvexLinear'):
            self.param.child('solver').show()
            self.param.child('tries_to_fix').show()
            self.param.child('tol').show()
        elif method in ('LocalizedHull', 'LinearSolver'):
            self.param.child('solver').hide()
            self.param.child('tries_to_fix').hide()
            self.param.child('tol').show()
        else:
            self.param.child('solver').hide()
            self.param.child('tries_to_fix').hide()
            self.param.child('tol').hide()
        
#        if method == 'LinearSolver':
#           self.param.child('force_update').setReadonly()
#           self.param.child('force_update').setValue(False)
#        else:
#           self.param.child('force_update').setReadonly(False)
        
        
    def setup_CS(self):
        sample_box = self.w.get_sample_box()
        sample_space = getattr(sample_box, self.w.space)
        
        #č kvůli Qhull se může stát, že budeme muset zachovat starý CS
        #č proto opatrně mažeme kwargs no a tak.
        method = self.param.child('method').value()
        if method == 'Qhull':
            status, qframe = self.get_qhull()
            if status:
                self.CS = qframe
                self.kwargs.clear()
            else:
                self.param.child('method').setToDefault()
        elif method == 'DirectContact':
            self.CS = wireframe.DirectContact(sample_space)
            self.kwargs.clear()
        elif method == 'Gabriel':
            self.CS = wireframe.Gabriel(sample_space)
            self.kwargs.clear()
        elif method == 'ConvexSolver':
            solver = getattr(wireframe, self.param.child('solver').value())
            self.CS = wireframe.ConvexSolver(sample_space, convex_solver=solver)
            #self.kwargs.clear() #č aby se to neztratilo až bude 100500 metodů
            self.kwargs['tries_to_fix'] = self.param.getValues()['tries_to_fix'][0]
            self.kwargs['tol'] = self.param.getValues()['tol'][0]
        elif method == 'ConvexLinear':
            solver = getattr(wireframe, self.param.child('solver').value())
            self.CS = wireframe.ConvexLinear(sample_space, convex_solver=solver)
            #self.kwargs.clear() #č aby se to neztratilo až bude 100500 metodů
            self.kwargs['tries_to_fix'] = self.param.getValues()['tries_to_fix'][0]
            self.kwargs['tol'] = self.param.getValues()['tol'][0]
        elif method == 'LinearSolver':
            self.CS = wireframe.LinearSolver(sample_space)
            self.kwargs.clear()
            self.kwargs['tol'] = self.param.getValues()['tol'][0]
        elif method == 'LocalizedHull':
            self.CS = wireframe.LocalizedHull(sample_space)
            self.kwargs.clear()
            self.kwargs['tol'] = self.param.getValues()['tol'][0]
        else:
            raise ValueError("ContactWidget: unknown method")
            
        
        




#!/usr/bin/env python
# coding: utf-8

# Axes level functions (for Matplotlib)

#č Založme nový modul pro funkce (procedury),
#č které sice stejně jako mart pracují na urovni axes,
#č ale na rozdil od atomických funkcí mart modulu
#č udělaj z prazdných os hotový obrazek.
#č mplot.show2D() může použivat tuhle nabídku obrázků.
#
#č funkce v tomto modulu dostávájí jako parameter jedině ax
#č ax ale má nastavené atributy .space a .sample_box

import numpy as np
from . import mart
from . import mgraph

from matplotlib import colors as mcolors
from matplotlib import ticker
from matplotlib import patches as mpatches
from matplotlib import lines as mlines


# it is mostly for qt_plot, it offers availiable options to user
__all__ =   [
            'tri_vector_plot',
            'GRaph',
            'rbf_plot', 'rbf_density_plot', 'rbf_diagram',
            'candidates_plot', 'rejection_sampling_plot', 
            'candidates_sampling_plot',
            'convex_hull_plot', 'tri_plot', 'tri_nodes_plot',
            'tri_R_plot', 'tri_GK_plot', 'tri_G_plot',
            'tri_R_nodes_plot', 'tri_GK_nodes_plot',
            'convergence_diagram', 'convergence_legend', 
            'convergence_square', 'beta_diagram',
            'just_points', 'just_points_really',
            'base_drawing', 'candidates_drawing',
            'just_qhull', 'qhull_plot', 'qhull_infinite',
            'dhull_scheme_plot', 'dhull_random_plot',
            'bhull_plot', 'bhull_infinite',
            'shull_plot',
            'completehull_plot'
            ]

#č ten [plot] zásadně vytvaří své struktury, nepouzívá oné ze skříňky,
#č protože já vím, že v těch obrázcích, ve kterých chcu ho použit,
#č můde být třeba použit řez a skříňka tedy potřebné struktury může nemít
def tri_vector_plot(ax, tri_space=None, linewidths=[0.7, 0.5, 0.4, 0.3, 0.2, 0.1], 
                ms=3, mew=0.6, lw=0.7, data_offset=1, nrid=200):
    from .. import simplex as six
    if tri_space is None:
        tri_space = ax.space
        
    import quadpy
    tn_scheme = quadpy.tn.stroud_tn_3_6b(2)
            
    Tri = six.GaussCubatureIntegration(ax.sample_box, tn_scheme, incremental=False)
    Tri.get_pf_estimation()
    ax.sample_box.Tri = Tri
    mart.setup(ax)
    if tri_space == ax.space:
        #mart.triplot(ax, color="#B2B2B2", lw=lw/1.4, zorder=100)
        s = (1, 1, 1, 0) #'xkcd:light seafoam green' #a7ffb5
        f = '#fdc1c500' #(253/255, 193/255, 197/255, 1) # '#fdc1c5' #'xkcd: pale rose' # (#fdc1c5)
        m = '#FFF39A00'#(255/255, 243/255, 154/255, 1)#'#FFF39A' #'xkcd: dark cream' # (255, 243, 154, 255)
        sfm_colors = [s, f, m]
        mart.tripcolor(ax, sfm_colors=sfm_colors, color="#B2B2B200", 
                                    linewidth=lw/1.4/2, zorder=100)
    else:
        mart.tri_plot(ax, Tri=Tri, color="#B2B2B2", lw=lw/1.4, zorder=100, rasterized=True)
    
    mart.simplex_vectors(ax, color='grey', zorder=10500)#, scale_units='height', scale=0.0025) #width=0.003,
    
    mart.plot_points(ax, ms=ms, mew=mew, zorder=100500)
    #č curly a boundaries raději nakreslíme co nejpozději
    mart.curly(ax, nrid=nrid, linewidths=linewidths)
    try:
        mart.plot_boundaries(ax, lw=lw, zorder=1050, nrod=nrid)
    except:
        pass
    min_coord = np.abs(np.min(getattr(ax.sample_box, ax.space)))
    mart.setup_labels(ax, data_offset=data_offset * min_coord)



def GRaph(ax, df=None, ls='', **kwargs):
    from .. import sball
    ax.margins(0)
    #ax.yaxis.set_ticks_position('right')
    #ax.yaxis.set_label_position('right')
    ax.set_xlabel(r"Limit-state evaluations, $N_{\mathrm{sim}}$")
    ax.set_ylabel(r"$\rho$")
    
    s_ball = sball.Sball(ax.sample_box.nvar)
    
    
    secax = ax.secondary_yaxis(1, functions=(s_ball.sf, s_ball.isf))
    secax.set_ylabel(r"$S_{\chi} (\rho; N_{\mathrm{var}})$")
    secax.yaxis.set_major_locator(ticker.LogLocator())
    secax.yaxis.set_major_formatter(ticker.LogFormatterMathtext())

    lengths = np.sum(np.square(ax.sample_box.G), axis=1)
    lengths = np.sqrt(lengths, out=lengths) #lengths of each radius-vector
    
    x = np.arange(len(lengths)) + 1
    y = lengths
    
    failsi = ax.sample_box.failsi
    
    try: # proxy denotes to implicitly-known values
        proxy = ax.sample_box.proxy.astype(bool) 
    except AttributeError:
        proxy = np.full(len(lengths), False, dtype=bool)
    
    
    #č byl jsem svědkem, že matplotlib zlobil ve 3D 
    #č kvůli tomu, že nebyl žádný safe vzorek
    #č proto raději budu přidávat tečky podmíněne
    mask = np.all((~failsi, ~proxy), axis=0)
    if np.any(mask): #success
        ax.plot(x[mask], y[mask], mec='g', mfc='g', marker='P', ls=ls, **kwargs)
    
    mask = np.all((failsi, ~proxy), axis=0)
    if np.any(mask): #failures
        ax.plot(x[mask], y[mask], mec='r', mfc='r', marker='X', ls=ls, **kwargs)
    
    mask = np.all((~failsi, proxy), axis=0)
    if np.any(mask): #proxy_successes
        ax.plot(x[mask], y[mask], 
                mec='#77AC30', mfc=(0,0,0,0), marker='h', ls=ls, **kwargs)
    
    mask = np.all((failsi, proxy), axis=0)
    if np.any(mask): #proxy_failures
        ax.plot(x[mask], y[mask], 
                mec='#D95319', mfc=(0,0,0,0), marker='H', ls=ls, **kwargs)
                
    
    if np.any(failsi): 
        ax.axhline(np.min(y[failsi]), c='r', zorder=-1, **kwargs)
    
    if df is None:
        try:
            import pandas as pd
            df = pd.DataFrame(ax.sample_box.box_estimations)
            df.index = df.nsim.to_numpy()
        except BaseException as e:
            print(repr(e))
            return 
        
    x = df.nsim.to_numpy()
    r = df.r.to_numpy()
    R = df.R.to_numpy()
    ax.fill_between(x, r, y2=R, color='#C3C3C3', step='post', zorder=-20)
    
    if 'r_safe' in df:
        y = df['r_safe'].to_numpy()
        y2 = df['R_safe'].to_numpy()
        mask = y >= 0
            
        ax.fill_between(x[mask], y[mask], y2[mask], color='#A7FFB5', step='post', zorder=-100)
    
    if 'r_failure' in df:
        y = df['r_failure'].to_numpy()
        y2 = df['R_failure'].to_numpy()
        mask = y >= 0
        
        ax.fill_between(x[mask], y[mask], y2[mask], color='#FDC1C5', step='post', zorder=-50)
        
    if 'r_mixed' in df:
        y = df['r_mixed'].to_numpy()
        y2 = df['R_mixed'].to_numpy()
        mask = y >= 0
            
        ax.fill_between(x[mask], y[mask], y2[mask], color='#FFF39A', step='post', zorder=-30)
    
    
    
    if 'r_exact' in dir(ax.sample_box): 
        ax.axhline(ax.sample_box.r_exact, c='b', zorder=-2, **kwargs)
    
    
    ax.set_facecolor('#E1E1E1')
    
    
    
    


def rbf_diagram(ax):
    from .. import misc as wmisc
    rbf_stm = []
    x = list(range(2, ax.sample_box.nsim))
    for i in x:
        rbf = wmisc.RBF_surrogate(ax.sample_box[:i])
        rbf_stm.append(rbf.get_pf_estimation(2000000))
        print(i, repr(np.array(rbf_stm)))
    
    
    ax.plot(np.array(x), np.array(rbf_stm))
    
    

def rbf_plot(ax, nrid=500):
    tri_plot(ax)
    x_bound = ax.get_xbound()
    y_bound = ax.get_ybound()
    mart.rbf_colormesh(ax, nrid)
    ax.set_xbound(x_bound)
    ax.set_ybound(y_bound)

def rbf_density_plot(ax, nrid=500):
    tri_plot(ax)
    x_bound = ax.get_xbound()
    y_bound = ax.get_ybound()
    mart.rbf_density_colormesh(ax, nrid)
    ax.set_xbound(x_bound)
    ax.set_ybound(y_bound)

def rejection_sampling_plot(ax, linewidths=[0.7, 0.5, 0.4, 0.3, 0.2, 0.1]):
    from .. import simplex as six
    
    #č možná zrovna zde nebudem robiť vědu a použijme tringulaciju ze skříňky?
    nodes = ax.sample_box.f_model(1000)
    # -1 = 'out', 0=success, 1=failure, 2=mix
    event_ids = six.filter(ax.sample_box.Tri, nodes)
    mart.plot_sample(ax, nodes[event_ids==-1], ls='', marker='.',\
                         mec="#00007E", mfc="#00007E", ms=1.5, alpha=0.5,\
                         rasterized=True)
    mart.plot_sample(ax, nodes[event_ids==2], ls='', marker='.',\
                         mec="#00007E", mfc="#00007E", ms=1.5, alpha=0.5,\
                         rasterized=True)
    
    mart.setup(ax)
    mart.curly(ax, linewidths=linewidths)
    mart.triplot(ax, color="#B2B2B2", lw=0.5, zorder=1000)
    mart.plot_points(ax, ms=5, zorder=100500)
    try:
        mart.plot_boundaries(ax, lw=0.7, zorder=10500)
    except:
        pass
    ax.xaxis.set_major_formatter(ticker.NullFormatter())
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
    ax.xaxis.set_major_locator(ticker.NullLocator())
    ax.yaxis.set_major_locator(ticker.NullLocator())


def candidates_sampling_plot(ax, linewidths=[0.7, 0.5, 0.4, 0.3, 0.2, 0.1]):
    from .. import simplex as six
    from .. import estimation as stm
    
    def _stm_draw_nodes(*args, **kwargs):
        #č v tomto plot bude šilený bordel v tom, kdo kolbek spouští
        #č a kdo mu co tam posílá. Na event a na nodes spolehat ale můžeme
        #
        # callback's signature: sx, indices=, simplex=, nodes=, cell_stats=
        # positional "sx" is Tri object itself
        # "indices" are numbers of simplex vertices
        # "simplex" are vertices itself
        # "nodes" is what we really want to draw
        
        event = kwargs['cell_stats']['event']
        
        if event in ('mix', 'outside'):
            mart.plot_sample(ax, kwargs['nodes'], ls='', marker='.',\
                         mec="#00007E", mfc="#00007E", ms=1.5, alpha=0.5,\
                         rasterized=True)
        
    data = stm.fast_simplex_estimation(ax.sample_box, model_space=ax.space,\
                                  sampling_space=ax.space, \
                                   weighting_space=ax.space,\
                                    outside_budget=1000, \
                                     simplex_budget=100,\
                                    callback=_stm_draw_nodes, design=None)    
    
    mart.setup(ax)
    mart.curly(ax, linewidths=linewidths)
    mart.triplot(ax, color="#B2B2B2", lw=0.5, zorder=1000)
    mart.plot_points(ax, ms=5, zorder=100500)
    try:
        mart.plot_boundaries(ax, lw=0.7, zorder=10500)
    except:
        pass
    ax.xaxis.set_major_formatter(ticker.NullFormatter())
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
    ax.xaxis.set_major_locator(ticker.NullLocator())
    ax.yaxis.set_major_locator(ticker.NullLocator())

def candidates_plot(ax):
    tri_plot(ax, tri_space=None, linewidths=[0.7, 0.5, 0.4, 0.3], data_offset=0.4)
    blue_colors = ["#BDBDFF", "#9999FF", "#00007E", "#000057"]
    blue_cmap = mcolors.LinearSegmentedColormap.from_list("bluecmap", blue_colors)
    mart.scatter_candidates(ax, s=1.5, marker='.', cmap=blue_cmap, rasterized=True)
    mart.plot_the_best_candidate(ax, "^", color='#000057')
    mart.plot_points(ax, ms=5, zorder=1005000)
    
#    if ax.space in ('G', 'GK'):
#        ax.xaxis.set_label_coords(1, 0.47)
#        ax.set_xlabel("$x_1$")
#        ax.text(0.47, 1, '$x_2$', ha='right',va='top', transform=ax.transAxes)
#    else:
#        ax.set_xlabel("$x_1$")
#        ax.set_ylabel("$x_2$")

def convergence_square(ax):
    convergence_diagram(ax)
    ax.set_box_aspect(1)
    ax.secondary_yaxis('right')

def convergence_diagram(ax):
    #č pokorně jedeme použiť guessbox
    #č nic jiného nebylo pořádně implementováno
    #from .. import stm_df 
    #df = stm_df.get_tri_data_frame(ax.sample_box, sources, apply_proxy)
    import pandas as pd
    df = pd.DataFrame(ax.sample_box.box_estimations)
    df.index = df.nsim.to_numpy()
    try:
        pf_exact = ax.sample_box.pf_exact
        pf_exact_method = ax.sample_box.pf_exact_method
        mgraph.tri_estimation_plot(ax, df, pf_exact=pf_exact, \
                            pf_exact_method=pf_exact_method, plot_outside=True)
    except:
        mgraph.tri_estimation_plot(ax, df, plot_outside=True)
    ax.margins(0)
    ax.set_yscale('log')
    ax.set_xlabel(r"Limit-state evaluations, $N_{\mathrm{sim}}$")
    #ax.set_xlabel("Number of points")
    ax.set_ylabel("Probability measure")


def beta_diagram(ax, df=None):
    if df is None:
        try:
            import pandas as pd
            df = pd.DataFrame(ax.sample_box.box_estimations)
            df.index = df.nsim.to_numpy()
        except BaseException as e:
            print(repr(e))
            return 
    try:
        pf_exact = ax.sample_box.pf_exact
        pf_exact_method = ax.sample_box.pf_exact_method
        mgraph.tri_beta_plot(ax, df, pf_exact=pf_exact, \
                            pf_exact_method=pf_exact_method, plot_outside=True)
    except:
        mgraph.tri_beta_plot(ax, df, plot_outside=True)
    ax.margins(0)
    ax.set_xlabel(r"Limit-state evaluations, $N_{\mathrm{sim}}$")
    #ax.set_xlabel("Number of points")
    ax.set_ylabel(r"Reliability index, $\beta$")

    
def convergence_legend(ax):
    convergence_diagram(ax)
    
    #č nejsem jist, jestli by šlo přehodit pořadí u stackplotu
    #č proto vytvařením legendy už neotravujeme mgraph
    #č ale sámi tu něco nahodíme
    proxy_handles = list()
    proxy_handles.append(mpatches.Patch(color='#A7FFB5', label=r"$\pazocal{S}^{\left(N_{\mathrm{sim}}\right)}$"))
    proxy_handles.append(mpatches.Patch(color='#DDDDDD', label=r"$\pazocal{O}^{\left(N_{\mathrm{sim}}\right)}$"))
    proxy_handles.append(mpatches.Patch(color='#FFF39A', label=r"$\pazocal{M}^{\left(N_{\mathrm{sim}}\right)}$"))
    proxy_handles.append(mpatches.Patch(color='#FDC1C5', label=r"$\pazocal{F}^{\left(N_{\mathrm{sim}}\right)}$"))
    
    proxy_handles.append(mlines.Line2D([], [], color='b', label=r'$p_{\pazocal{F}}$'))
    #proxy_handles.append(mlines.Line2D([], [], color='m', label=r"$p_{\mathrm{f,v}}^{\left(N_{\mathrm{sim}}\right)}$"))
    proxy_handles.append(mlines.Line2D([], [], color='r', label=r"$p_{\mathrm{f,v}}^{\left(N_{\mathrm{sim}}\right)}$"))
    proxy_handles.append(mlines.Line2D([], [], color='#AAAAAA', \
                label=r"$p_{\pazocal{O}}^{\left(N_{\mathrm{sim}}\right)}$"))
    
    ax.set_xlabel(r"Number of $g\left( \bm{x} \right)$ evaluations, $N_{\mathrm{sim}}$")
    #ax.set_box_aspect(0.5)
    ax.legend(bbox_to_anchor=(1.02, 1), loc='upper left', handles=proxy_handles, frameon=True, borderaxespad=0,
                handlelength=1, handletextpad = 0.6)
    #             frameon=True)#, mode="expand")
    #ax.legend(bbox_to_anchor=(0.5, -0.25), ncol=2, loc='upper center', handles=proxy_handles)


#č ten [plot] zásadně vytvaří své struktury, nepouzívá oné ze skříňky,
#č protože já vím, že v těch obrázcích, ve kterých chcu ho použit,
#č můde být třeba použit řez a skříňka tedy potřebné struktury může nemít
def tri_nodes_plot(ax, tri_space=None, tn_scheme=None, ms=3, mew=0.6, lw=0.7,
                     linewidths=[0.7, 0.5, 0.4, 0.3, 0.2, 0.1], nrid=200,
                     data_offset=1):
    from .. import simplex as six
    if tri_space is None:
        tri_space = ax.space
    
    #č já tuhle funkciju potřebuju ne abych kreslil bodíky ve vrcholech
    if tn_scheme is None:
        try:
            tn_scheme = ax.sample_box.Tri.tn_scheme
        except:
            import quadpy
            tn_scheme = quadpy.tn.grundmann_moeller(ax.sample_box.nvar, 5)
            
    
    def _draw_nodes(*args, **kwargs):
        # callback's signature: sx, indices=, simplex=, nodes=, cell_stats=
        # positional "sx" is Tri object itself
        # "indices" are numbers of simplex vertices
        # "simplex" are vertices itself
        # "nodes" is what we really want to draw
        
        event = kwargs['cell_stats']['event']
        
        if event == 'mix':
            color = '#FFF39A' #'xkcd: dark cream' # (255, 243, 154, 255)
        elif event == 'failure':
            color = '#fdc1c5' #'xkcd: pale rose' # (#fdc1c5)
        elif event == 'success':
            color = '#a7ffb5' #'xkcd:light seafoam green' #a7ffb5
        else:
            assert 100500 < 0 #оӵ мар лэсьтӥське татын?
            
        mart.plot_sample(ax, kwargs['nodes'], ls='', marker='.',\
                         mew=0, mfc=color, ms=2, alpha=0.5, rasterized=True)
        
        
        
    
    #č vytvařím vlastní tringulaciju zde vécemeně kvůli callbackům
    #č jde, samozřejmě všecko udělat jínak, 
    #č ale nechcu zrovna zde z toho robiť vědu
    Tri = six.JustCubatureTriangulation(ax.sample_box, tn_scheme=tn_scheme, \
                                    tri_space=tri_space, issi=None,\
                                    weighting_space=None, \
                                    incremental=False,\
                                    on_add_simplex=_draw_nodes,\
                                    on_delete_simplex=None)
    
    mart.setup(ax)
    Tri.integrate()
    if tri_space == ax.space:
        mart.triplot(ax, color="#B2B2B2", lw=lw/1.4, zorder=100)
    else:
        mart.tri_plot(ax, Tri=Tri, color="#B2B2B2", lw=lw/1.4, zorder=100, rasterized=True)
    
    mart.plot_points(ax, ms=ms, mew=mew, zorder=100500)
    #č curly a boundaries raději nakreslíme co nejpozději
    mart.curly(ax, nrid=nrid, linewidths=linewidths)
    try:
        mart.plot_boundaries(ax, lw=lw, zorder=1050, nrod=nrid)
    except:
        pass
    min_coord = np.abs(np.min(getattr(ax.sample_box, ax.space)))
    mart.setup_labels(ax, data_offset=data_offset * min_coord)

def tri_R_nodes_plot(ax, **kwargs):
    tri_nodes_plot(ax, tri_space='R', **kwargs)
    
def tri_G_plot(ax, **kwargs):
    tri_plot(ax, tri_space='G', **kwargs)
    
def tri_GK_nodes_plot(ax, **kwargs):
    tri_nodes_plot(ax, tri_space='GK', **kwargs)


#č ten [plot] zásadně vytvaří své struktury, nepouzívá oné ze skříňky,
#č protože já vím, že v těch obrázcích, ve kterých chcu ho použit,
#č můde být třeba použit řez a skříňka tedy potřebné struktury může nemít
def tri_plot(ax, tri_space=None, linewidths=[0.7, 0.5, 0.4, 0.3, 0.2, 0.1], 
                ms=3, mew=0.6, lw=0.7, data_offset=1, nrid=200):
    from .. import simplex as six
    if tri_space is None:
        tri_space = ax.space
    
    Tri = six.JustCubatureTriangulation(ax.sample_box, tn_scheme=None, \
                                    tri_space=tri_space, issi=None,\
                                    weighting_space=None, \
                                    incremental=False,\
                                    on_add_simplex=None,\
                                    on_delete_simplex=None)
    
    mart.setup(ax)
    if tri_space == ax.space:
        mart.triplot(ax, color="#B2B2B2", lw=lw/1.4, zorder=100)
    else:
        mart.tri_plot(ax, Tri=Tri, color="#B2B2B2", lw=lw/1.4, zorder=100, rasterized=True)
    
    mart.plot_points(ax, ms=ms, mew=mew, zorder=100500)
    #č curly a boundaries raději nakreslíme co nejpozději
    mart.curly(ax, nrid=nrid, linewidths=linewidths)
    try:
        mart.plot_boundaries(ax, lw=lw, zorder=1050, nrod=nrid)
    except:
        pass
    min_coord = np.abs(np.min(getattr(ax.sample_box, ax.space)))
    mart.setup_labels(ax, data_offset=data_offset * min_coord)
        
    

def tri_R_plot(ax, **kwargs):
    tri_plot(ax, tri_space='R', **kwargs)
    
def tri_GK_plot(ax, **kwargs):
    tri_plot(ax, tri_space='GK', **kwargs)

#č ten [plot] zásadně vytvaří svou obálku, nepouzívá onou ze skříňky,
#č protože já vím, že v těch obrázcích, ve kterých chcu ho použit,
#č můde být třeba použit řez a skříňka tedy potřebné struktury může nemít
def convex_hull_plot(ax, tri_space=None, ms=3, mew=0.6, lw=0.7, nrid=200,
                    linewidths=[0.7, 0.5, 0.4, 0.3, 0.2, 0.1]):
    from .. import convex_hull as khull
    if tri_space is None:
        tri_space = ax.space
    
    mart.setup(ax)
    qhull = khull.QHull(ax.sample_box, space=tri_space, incremental=False)
    mart.qhull_plot(ax, qhull, color="#B2B2B2", lw=lw, zorder=100)
    mart.plot_points(ax, ms=ms, mew=mew, zorder=100500)
    #č curly a boundaries raději nakreslíme co nejpozději
    mart.curly(ax, nrid=nrid, linewidths=linewidths)
    try:
        mart.plot_boundaries(ax, lw=lw, zorder=1050, nrod=nrid)
    except:
        pass
    mart.setup_labels(ax)

def just_points(ax):
    ax.set_xlabel('$x_{1}$')
    ax.set_ylabel('$x_{2}$')
    
    ax.set_aspect(1)
    #ax.set_box_aspect(1)
    mart.scatter_points(ax)
    
def just_points_really(ax):
    ax.set_aspect(1)
    ax.set_frame_on(False)
    mart.scatter_points(ax)

def base_drawing(ax):
    mart.setup(ax)
    mart.curly(ax)
    try:
        mart.triplot(ax, color="grey", linewidth=0.4)
    except:
        pass
    mart.plot_boundaries(ax, linewidth=0.7)
    mart.scatter_points(ax)
    
    
def candidates_drawing(ax):
    try:
        mart.triplot(ax, color="grey", linewidth=0.4)
    except:
        pass
    mart.plot_boundaries(ax, linewidth=0.7)
    
    #č ax.scatter posílá parameter cmap Collections třídě.
    #č Třída mimo jiného dědí cm.ScalarMappable,
    #č která inicializaci deleguje funkci cm.get_cmap() ve (svém) modulu cm.
    # cmap='viridis_r' #cmap='plasma',
    mart.scatter_candidates(ax, s=5, marker='.', cmap='plasma_r',\
     alpha=None, linewidths=None, edgecolors=None, plotnonfinite=False, rasterized=True)
    mart.plot_the_best_candidate(ax, "^", color='#3D0D5B')
    
        
    
    

# defaults
hezkymodře = (85/255, 70/255, 1, 1)
inside_color = [(185/255, 228/255, 14/255, 1)]
outside_color = [(128/255, 128/255, 128/255, 0.5)]
hull_colors = (hezkymodře, inside_color, outside_color) # border_inside_outside

def just_qhull(ax, lw=1.5, **kwargs):
    from .. import convex_hull as khull
    qhull = khull.QHull(ax.sample_box, space=ax.space, incremental=False)
    border_color, _, __ = hull_colors

    # setup
    ax.set_aspect(1)
    #ax.set_box_aspect(3/4)
    #ax.set_xlim(-lim, lim)
    #ax.set_ylim(-lim, lim)
    ax.set_xlabel('$x_1$')
    ax.set_ylabel('$x_2$')
    
    mart.qhull_plot(ax, qhull, color=border_color, lw=lw, zorder=100, **kwargs)
    
    # finally samples
    #
    #mart.scatter_sample(ax, f, c='g', marker='P', zorder=1000) # why?
    mart.scatter_points(ax, zorder=100500)



def _hull_model_plot(ax, hull, hull_plot, ns=50000, lim=3, lw=1.5, s=4, hull_colors=hull_colors, **kwargs):
    border_color, inside_color, outside_color = hull_colors

    # setup
    ax.set_aspect(1)
    #ax.set_box_aspect(3/4)
    ax.set_xlim(-lim, lim)
    ax.set_ylim(-lim, lim)
    ax.set_xlabel('$x_1$')
    ax.set_ylabel('$x_2$')
    
    nodes = ax.sample_box.f_model(ns)
    mask = hull.is_outside(nodes)
    mart.scatter_sample(ax, nodes[mask], c=outside_color, marker='.', s=s)
    mart.scatter_sample(ax, nodes[~mask], c=inside_color, marker='.', s=s)
    hull_plot(ax, hull, color=border_color, lw=lw, zorder=100, **kwargs)
    
    # finally samples
    #
    #mart.scatter_sample(ax, f, c='g', marker='P', zorder=1000) # why?
    mart.scatter_points(ax, zorder=100500)
    

    
    
    
def qhull_plot(ax, **kwargs):
    from .. import convex_hull as khull
    qhull = khull.QHull(ax.sample_box, space=ax.space, incremental=False)
    _hull_model_plot(ax, qhull, mart.qhull_plot, **kwargs)
    
    #mart.qhull_polygon(ax, qhull, fc="white", ec="black", lw=0.75)


def qhull_infinite(ax, **kwargs):
    from .. import convex_hull as khull
    qhull = khull.QHull(ax.sample_box, space=ax.space, incremental=False)
    _hull_model_plot(ax, qhull, mart.dhull_plot, **kwargs)




def dhull_scheme_plot(ax, **kwargs):
    from .. import convex_hull as khull
    import quadpy
    
    scheme = quadpy.un.stroud_un_7_1(2)
    dhull = khull.DirectHull(ax.sample_box, scheme.points, space=ax.space)
    _hull_model_plot(ax, dhull, mart.dhull_plot, **kwargs)
    

def dhull_random_plot(ax, ndir=8, rand_dir=None, **kwargs):
    from .. import convex_hull as khull
    from .. import sball
    
    if rand_dir is None:
        rand_dir = sball.get_random_directions(ndir, 2)
    dhull = khull.DirectHull(ax.sample_box, rand_dir, space=ax.space)
    _hull_model_plot(ax, dhull, mart.dhull_plot, **kwargs)  
    
    
def bhull_plot(ax, **kwargs):
    from .. import convex_hull as khull
    
    bhull = khull.BrickHull(ax.sample_box, space=ax.space)
    _hull_model_plot(ax, bhull, mart.bhull_plot, **kwargs)
    
def bhull_infinite(ax, **kwargs):
    from .. import convex_hull as khull
    
    bhull = khull.BrickHull(ax.sample_box, space=ax.space)
    _hull_model_plot(ax, bhull, mart.dhull_plot, **kwargs)    
    
    
def shull_plot(ax, **kwargs):
    from .. import convex_hull as khull
    
    shull = khull.GBall(ax.sample_box)
    _hull_model_plot(ax, shull, mart.shull_plot, **kwargs)
    
    

def completehull_plot(ax, ndir=3, rand_dir=None, **kwargs):
    from .. import convex_hull as khull
    from .. import sball
    
    if rand_dir is None:
        rand_dir = sball.get_random_directions(ndir, 2)
    
    copletehull = khull.CompleteHull(ax.sample_box, rand_dir, space=ax.space)
    shull = khull.GBall(ax.sample_box)
    #_hull_model_plot(ax, copletehull, mart.dhull_plot, **kwargs)
    #_hull_model_plot(ax, copletehull, mart.shull_plot, ns=0, **kwargs)
    
    ns=50000
    lim=3
    lw=1.5
    s=4
    #hull_colors=hull_colors, **kwargs):
    border_color, inside_color, outside_color = hull_colors

    # setup
    ax.set_aspect(1)
    #ax.set_box_aspect(3/4)
    ax.set_xlim(-lim, lim)
    ax.set_ylim(-lim, lim)
    ax.set_xlabel('$x_1$')
    ax.set_ylabel('$x_2$')
    
    nodes = ax.sample_box.f_model(ns)
    completemask = copletehull.is_outside(nodes)
    smask = shull.is_outside(nodes)
    mask = np.any([completemask, smask], axis=0)
    mart.scatter_sample(ax, nodes[mask], c=outside_color, marker='.', s=s)
    mart.scatter_sample(ax, nodes[~mask], c=inside_color, marker='.', s=s)
    mart.dhull_plot(ax, copletehull, color=border_color, lw=lw, zorder=100, **kwargs)
    mart.shull_plot(ax, shull, color=border_color, lw=lw/2, ls='--', zorder=100, **kwargs)
    
    # finally samples
    #
    #mart.scatter_sample(ax, f, c='g', marker='P', zorder=1000) # why?
    mart.scatter_points(ax, zorder=100500)
    



#!/usr/bin/env python
# coding: utf-8

import matplotlib


__all__ = ['margins3d_patch', 'grid3d_patch', 'enable_tex']

def margins3d_patch(*args, **kwargs):
    from . import _axis3d_margins_patch
    
def grid3d_patch(*args, **kwargs):
    from . import _axis3d_margins_patch, _axes3d


def enable_tex(*args, **kwargs):
    matplotlib.rcParams['text.usetex'] = True
    
    # Times, Palatino, New Century Schoolbook, Bookman, Computer Modern Roman
    matplotlib.rc('font',**{'family':'serif','serif':['Computer Modern Roman']}) 
    #matplotlib.use('pgf')
    #matplotlib.rcParams["pgf.texsystem"] = "xelatex"
    #matplotlib.rcParams["pgf.rcfonts"] = False
    
    preamble = r'''\usepackage[utf8]{inputenc} %unicode support
    \usepackage[T1]{fontenc}
    \DeclareMathAlphabet{\pazocal}{OMS}{zplm}{m}{n}
    \usepackage{calrsfs}
    \usepackage{amsmath}
    \usepackage{bm}
    \usepackage[bitstream-charter]{mathdesign}
    '''
    
    
    #with open("preamble.tex", "r") as file:
    #    preamble = file.read()
    
    matplotlib.rc('text.latex', preamble=preamble)
    #matplotlib.rcParams["pgf.preamble"] = preamble



#!/usr/bin/env python
# coding: utf-8

import quadpy

un_spheres = [
    "dobrodeev_1978",
    "mysovskikh_1",
    "mysovskikh_2",
    "stroud_un_3_1",
    "stroud_un_3_2",
    "stroud_un_5_1",
    "stroud_un_5_2",
    "stroud_un_5_3",
    "stroud_un_5_4",
    "stroud_un_7_2",
    "stroud_un_7_1",
    "stroud_un_11_1",
#    "stroud_1967", # dup of "stroud_un_7_1"
#    "stroud_1969", # dup of "stroud_un_11_1"
]

#č Bacha! Lidstvo nemusí nutně dočkat 
#č ukončení výpočtu této funkce
def get_all_un_sphere_schemes(dim):
    schemes = dict()
    for item in un_spheres:
        Scheme = getattr(quadpy.un, item)
        try:
            scheme = Scheme(dim)
            schemes[scheme.name] = scheme
        except:
            pass
            
    return schemes



# up to degree 7. Ought to be enough for anybody
#č 7 stupňů musí stačit každému
def get_all_tn_simplex_schemes(dim):
    #č současný qt_plot vytahuje schemata ze slovníku
    #č aspoň netřeba v dlouhém seznamu dohledávat 
    schemes = dict()
    # Grundmann, Möller
    for i in range(8):
        scheme = quadpy.tn.grundmann_moeller(dim, i)
        schemes[scheme.name] = scheme
    # Silvester
    for i in range(7):
        scheme = quadpy.tn.silvester(dim, variant="open", n=i+1, symbolic=False)
        schemes[scheme.name] = scheme
        scheme = quadpy.tn.silvester(dim, variant="closed", n=i+1, symbolic=False)
        schemes[scheme.name] = scheme
    
    # Laufer, Hammer and Stroud
    for item in tn_simplices:
        Scheme = getattr(quadpy.tn, item)
        try:
            scheme = Scheme(dim)
            schemes[scheme.name] = scheme
        except:
            print(item, "Excluded")
        
    return schemes
    
    # v - has points at vertices
    # i - scheme has not integration points at vertices
    #
    # p - positive, all weights are positive
    # n - scheme has negative weights as well
tn_simplices = [
                'stroud_tn_1_1', # i p # single centroid point
                'stroud_tn_1_2', # v p # vertices only # laufer_1
                'stroud_tn_2_1a', # i p, positive, stejné váhy # ndim + 1 points # Hammer-Stround 1a # umísuje bodíky s odstupem od středu směrem k vrcholům
                'stroud_tn_2_1b', # i p, positive, stejné váhy # ndim + 1 points # Hammer-Stround 1b # umísuje bodíky na stěny simplexu 
                'stroud_tn_2_2', # negative # laufer_2
                'stroud_tn_3_1', # i negative
                'stroud_tn_3_2', # v negative
                'stroud_tn_3_3', # v negative pro ndim >=4
                'stroud_tn_3_4', # i p, positive # 45 points at 8D # Stroud 1966-II
                'stroud_tn_3_5', # v negative pro ndim >=7
                'stroud_tn_3_6a', # i p, positive, stejné váhy # 72 points at 8D # Stroud 1964a # obecně ndim <= 9 
                'stroud_tn_3_6b', # i p, positive, stejné váhy # 72 points at 8D # Stroud 1964b
                'stroud_tn_3_7', # v, positive pro ndim <= 9 # 94 points at 8D
                'stroud_tn_3_8', # v, positive pro ndim <= 9 # 102 points at 8D
                'stroud_tn_3_9', # v n # laufer_3
                'stroud_tn_3_10', # i n (depends on dimension)
                'stroud_tn_3_11', # v n (depends on dimension)
                'stroud_tn_4_1', # v negative # laufer_4 n>=3
                'stroud_tn_5_1', # i n
                'stroud_tn_5_2', # v negative # laufer_5 n>=4
                'walkington_1', # i p # single centroid point
                'walkington_2', # bůhvíco, positive # I have strong suspicious, that it places points outside the simplex
                'walkington_3', # i negative
                'walkington_5', #owntest n=2,3 
                'walkington_7' #owntest n=3
                ]

tn_simplices_2d = [
                'stroud_tn_1_1', 
                'stroud_tn_1_2', # laufer_1
                'stroud_tn_2_1a', 
                'stroud_tn_2_1b', 
                'stroud_tn_2_2', # laufer_2
                'stroud_tn_3_1', 
                'stroud_tn_3_2', 
                'stroud_tn_3_3', 
                'stroud_tn_3_4', 
                'stroud_tn_3_6a', 
                'stroud_tn_3_6b',
                'stroud_tn_3_8', 
                'stroud_tn_3_9', # laufer_3 
                'walkington_1', 
                'walkington_2', 
                'walkington_3', 
                'walkington_5', #owntest n=2,3 
                ]
                
def get_t2_keys():
    return list(quadpy.t2.schemes.keys()).extend(tn_simplices_2d)
    
def get_t3_keys():
    return list(quadpy.t3.schemes.keys()).extend(tn_simplices)
    
def get_t2_scheme(key):
    try:
        return quadpy.t2.schemes[key]()
    except KeyError:
        Scheme = getattr(quadpy.tn, key)
        return Scheme(2)

def get_t3_scheme(key):
    try:
        return quadpy.t3.schemes[key]()
    except KeyError:
        Scheme = getattr(quadpy.tn, key)
        return Scheme(3)
        
        
def get_tn_keys(ndim):
    keys = ['Grundmann-Möller', 'Silvester open', 'Silvester closed']
    if ndim == 2:
        keys.extend(quadpy.t2.schemes.keys())
        keys.extend(tn_simplices_2d)
    elif ndim == 3:
        keys.extend(quadpy.t3.schemes.keys())
        keys.extend(tn_simplices)
        keys.remove('stroud_tn_5_2')
    elif ndim == 4:
        keys.extend(tn_simplices)
        keys.remove('walkington_5')
        keys.remove('walkington_7')
    elif ndim == 5:
        keys.extend(tn_simplices)
        keys.remove('stroud_tn_3_10')
        keys.remove('stroud_tn_3_11')
        keys.remove('walkington_5')
        keys.remove('walkington_7')
        
    else:
        keys.extend(tn_simplices)
        keys.remove('walkington_5')
        keys.remove('walkington_7')
    return keys


def get_tn_scheme(key, ndim, degree=5):
    if key == 'Grundmann-Möller': #č husté, ale záporné schema. 
        #č n=0 odpovidá jednomu bodíku v těžišti
        #č n=1 odpovidá ndim+1 bodům (šíkmě rovině)
        return quadpy.tn.grundmann_moeller(ndim, degree)
    elif key == 'Silvester open': #č husté positivní (pro degree 1) schema. 
        #č n=1 odpovidá odpovidá ndim+1 bodům (šíkmě rovině). Positivní.
        return quadpy.tn.silvester(ndim, variant="open", n=degree, symbolic=False)
    elif key == 'Silvester closed': #č na hranici, nic dovnitř nedává
        #č n=1 odpovídá bodíkům jen ve vrcholech
        return quadpy.tn.silvester(ndim, variant="closed", n=degree, symbolic=False)
    elif ndim == 2:
        return get_t2_scheme(key)
    elif ndim == 3:
        return get_t3_scheme(key)
    else:
        Scheme = getattr(quadpy.tn, key)
        return Scheme(ndim)
    
    
    
    
            



#!/usr/bin/env python
# coding: utf-8


# Monkey patch module
# Removes side panes on Axes3D plot
# Just import it to make MPL unpredictable

import numpy as np
from matplotlib import artist
from mpl_toolkits.mplot3d.axis3d import Axis
from mpl_toolkits.mplot3d.axis3d import art3d, proj3d, move_from_center, tick_update_position

from mpl_toolkits.mplot3d.axes3d import Axes3D
from matplotlib.axes import Axes


@artist.allow_rasterization
def draw(self, renderer):
    self.label._transform = self.axes.transData
    renderer.open_group('axis3d', gid=self.get_gid())

    ticks = self._update_ticks()

    info = self._axinfo
    index = info['i']

    mins, maxs, centers, deltas, tc, highs = self._get_coord_info(renderer)

    # Determine grid lines
    minmax = np.where(highs, maxs, mins)
    maxmin = np.where(highs, mins, maxs)

    # Draw main axis line
    juggled = info['juggled']
    edgep1 = minmax.copy()
    edgep1[juggled[0]] = maxmin[juggled[0]]

    edgep2 = edgep1.copy()
    edgep2[juggled[1]] = maxmin[juggled[1]]
    pep = np.asarray(
        proj3d.proj_trans_points([edgep1, edgep2], renderer.M))
    centpt = proj3d.proj_transform(*centers, renderer.M)
    self.line.set_data(pep[0], pep[1])
    self.line.draw(renderer)

    # Grid points where the planes meet
    xyz0 = np.tile(minmax, (len(ticks), 1))
    xyz0[:, index] = [tick.get_loc() for tick in ticks]

    # Draw labels
    # The transAxes transform is used because the Text object
    # rotates the text relative to the display coordinate system.
    # Therefore, if we want the labels to remain parallel to the
    # axis regardless of the aspect ratio, we need to convert the
    # edge points of the plane to display coordinates and calculate
    # an angle from that.
    # TODO: Maybe Text objects should handle this themselves?
    dx, dy = (self.axes.transAxes.transform([pep[0:2, 1]]) -
              self.axes.transAxes.transform([pep[0:2, 0]]))[0]

    lxyz = 0.5 * (edgep1 + edgep2)

    # A rough estimate; points are ambiguous since 3D plots rotate
    ax_scale = self.axes.bbox.size / self.figure.bbox.size
    ax_inches = np.multiply(ax_scale, self.figure.get_size_inches())
    ax_points_estimate = sum(72. * ax_inches)
    deltas_per_point = 48 / ax_points_estimate
    default_offset = 21.
    labeldeltas = (
        (self.labelpad + default_offset) * deltas_per_point * deltas)
    axmask = [True, True, True]
    axmask[index] = False
    lxyz = move_from_center(lxyz, centers, labeldeltas, axmask)
    tlx, tly, tlz = proj3d.proj_transform(*lxyz, renderer.M)
    self.label.set_position((tlx, tly))
    if self.get_rotate_label(self.label.get_text()):
        angle = art3d._norm_text_angle(np.rad2deg(np.arctan2(dy, dx)))
        self.label.set_rotation(angle)
    self.label.set_va(info['label']['va'])
    self.label.set_ha(info['label']['ha'])
    self.label.draw(renderer)

    # Draw Offset text

    # Which of the two edge points do we want to
    # use for locating the offset text?
    if juggled[2] == 2:
        outeredgep = edgep1
        outerindex = 0
    else:
        outeredgep = edgep2
        outerindex = 1

    pos = move_from_center(outeredgep, centers, labeldeltas, axmask)
    olx, oly, olz = proj3d.proj_transform(*pos, renderer.M)
    self.offsetText.set_text(self.major.formatter.get_offset())
    self.offsetText.set_position((olx, oly))
    angle = art3d._norm_text_angle(np.rad2deg(np.arctan2(dy, dx)))
    self.offsetText.set_rotation(angle)
    # Must set rotation mode to "anchor" so that
    # the alignment point is used as the "fulcrum" for rotation.
    self.offsetText.set_rotation_mode('anchor')

    #----------------------------------------------------------------------
    # Note: the following statement for determining the proper alignment of
    # the offset text. This was determined entirely by trial-and-error
    # and should not be in any way considered as "the way".  There are
    # still some edge cases where alignment is not quite right, but this
    # seems to be more of a geometry issue (in other words, I might be
    # using the wrong reference points).
    #
    # (TT, FF, TF, FT) are the shorthand for the tuple of
    #   (centpt[info['tickdir']] <= pep[info['tickdir'], outerindex],
    #    centpt[index] <= pep[index, outerindex])
    #
    # Three-letters (e.g., TFT, FTT) are short-hand for the array of bools
    # from the variable 'highs'.
    # ---------------------------------------------------------------------
    if centpt[info['tickdir']] > pep[info['tickdir'], outerindex]:
        # if FT and if highs has an even number of Trues
        if (centpt[index] <= pep[index, outerindex]
                and np.count_nonzero(highs) % 2 == 0):
            # Usually, this means align right, except for the FTT case,
            # in which offset for axis 1 and 2 are aligned left.
            if highs.tolist() == [False, True, True] and index in (1, 2):
                align = 'left'
            else:
                align = 'right'
        else:
            # The FF case
            align = 'left'
    else:
        # if TF and if highs has an even number of Trues
        if (centpt[index] > pep[index, outerindex]
                and np.count_nonzero(highs) % 2 == 0):
            # Usually mean align left, except if it is axis 2
            if index == 2:
                align = 'right'
            else:
                align = 'left'
        else:
            # The TT case
            align = 'right'

    self.offsetText.set_va('center')
    self.offsetText.set_ha(align)
    self.offsetText.draw(renderer)

    if self.axes._draw_grid and len(ticks):
        # Grid lines go from the end of one plane through the plane
        # intersection (at xyz0) to the end of the other plane.  The first
        # point (0) differs along dimension index-2 and the last (2) along
        # dimension index-1.
        lines = np.stack([xyz0, xyz0, xyz0], axis=1)
        
        lines[:, 0, index - 2] = maxmin[index - 2]
        lines[:, 2, index - 1] = maxmin[index - 1]
        
        lines[:, :, 2] = np.min(lines[:, :, 2])
        #print("lajny:", lines[:, :, 2])
        self.gridlines.set_segments(lines)
        self.gridlines.set_color(info['grid']['color'])
        self.gridlines.set_linewidth(info['grid']['linewidth'])
        self.gridlines.set_linestyle(info['grid']['linestyle'])
        self.gridlines.draw(renderer, project=True)

    # Draw ticks
    tickdir = info['tickdir']
    tickdelta = deltas[tickdir]
    if highs[tickdir]:
        ticksign = 1
    else:
        ticksign = -1

    for tick in ticks:
        # Get tick line positions
        pos = edgep1.copy()
        pos[index] = tick.get_loc()
        pos[tickdir] = (
            edgep1[tickdir]
            + info['tick']['outward_factor'] * ticksign * tickdelta)
        x1, y1, z1 = proj3d.proj_transform(*pos, renderer.M)
        pos[tickdir] = (
            edgep1[tickdir]
            - info['tick']['inward_factor'] * ticksign * tickdelta)
        x2, y2, z2 = proj3d.proj_transform(*pos, renderer.M)

        # Get position of label
        default_offset = 8.  # A rough estimate
        labeldeltas = (
            (tick.get_pad() + default_offset) * deltas_per_point * deltas)

        axmask = [True, True, True]
        axmask[index] = False
        pos[tickdir] = edgep1[tickdir]
        pos = move_from_center(pos, centers, labeldeltas, axmask)
        lx, ly, lz = proj3d.proj_transform(*pos, renderer.M)

        tick_update_position(tick, (x1, x2), (y1, y2), (lx, ly))
        tick.tick1line.set_linewidth(
            info['tick']['linewidth'][tick._major])
        tick.draw(renderer)

    renderer.close_group('axis3d')
    self.stale = False

Axis.draw = draw



@artist.allow_rasterization
def draw(self, renderer):
    # draw the background patch
    self.patch.draw(renderer)
    self._frameon = False

    # first, set the aspect
    # this is duplicated from `axes._base._AxesBase.draw`
    # but must be called before any of the artist are drawn as
    # it adjusts the view limits and the size of the bounding box
    # of the axes
    locator = self.get_axes_locator()
    if locator:
        pos = locator(self, renderer)
        self.apply_aspect(pos)
    else:
        self.apply_aspect()

    # add the projection matrix to the renderer
    self.M = self.get_proj()
    renderer.M = self.M
    renderer.vvec = self.vvec
    renderer.eye = self.eye
    renderer.get_axis_position = self.get_axis_position

    # Calculate projection of collections and patches and zorder them.
    # Make sure they are drawn above the grids.
    zorder_offset = max(axis.get_zorder()
                        for axis in self._get_axis_list()) + 1
    for i, col in enumerate(
            sorted(self.collections,
                   key=lambda col: col.do_3d_projection(renderer),
                   reverse=True)):
        col.zorder = zorder_offset + i
    for i, patch in enumerate(
            sorted(self.patches,
                   key=lambda patch: patch.do_3d_projection(renderer),
                   reverse=True)):
        patch.zorder = zorder_offset + i

    if self._axis3don:
        # Draw panes first
        #print(self._get_axis_list())
        axis_list = self._get_axis_list()
        #axis_list[0].draw_pane(renderer)
        #axis_list[1].draw_pane(renderer)
        #for axis in self._get_axis_list():
        #    axis.draw_pane(renderer)
        
        # Then axes
        #print(self._get_axis_list())
        #for axis in self._get_axis_list():
        #    axis.draw(renderer)
        axis_list[0].draw(renderer)
        axis_list[1].draw(renderer)


    # Then rest
    Axes.draw(self, renderer)

Axes3D.draw = draw





#!/usr/bin/env python
# coding: utf-8


import numpy as np
import quadpy

from ..ghull import Ghull
from .. import convex_hull as khull
from .. import simplex as sx
from ..reader import Store
from .. import sball

from scipy import stats
from scipy import spatial

from collections import namedtuple
from sortedcollections import ValueSortedDict



    
    



class _Exploration:
    
    outer_budget = 100
    psi_q = 0.5
        
    def __call__(bx):
        if bx.nsim < 1: # je to legální
            return bx.f_model.new_sample([], space='G', extend=True)
        elif bx.holydays and not bx.nsim % bx.holydays:
            bx._logger(msg='holyday!')
            return bx.holyday()
        elif bx.to_refine > bx.to_explore:
            bx._logger(msg='refine!')
            return bx.refine()
        else:
            return bx.explore()
            
            
            
    def explore(bx):
        r = bx.get_exploratory_radius()
        
        #č konkretně tato třída je pevně napojena na G prostor
        #č ale bacha, kbyby se to změnilo...
        assert bx.convex_hull.space == 'G'
        
        a, b, _sample = bx.convex_hull.get_exploration_vector()
        R = -b
            
        if r < R:
            bx._logger(msg='refine (fallbacked)!')
            return bx.refine()
        
        orth_nodes_T = np.random.randn(len(a), bx.outer_budget) # len(a) == ndim
        orth_basis = khull.get_orth_basis(a)
        sample_from = stats.norm.sf(r)
        t = np.linspace(sample_from, 0, bx.outer_budget, endpoint=False)
        orth_nodes_T[0] = stats.norm.isf(t)
            
        outside_nodes_G = (orth_basis.T @ orth_nodes_T).T
        outside_nodes = bx.f_model.new_sample(outside_nodes_G, space='G')
        
        tree = spatial.KDTree(bx.G, compact_nodes=True, balanced_tree=False)
        d1, i1 = tree.query(outside_nodes_G, k=1)
        
        PDF = bx.pdf('G')
        
        nodes_pdf = outside_nodes.pdf('G')
        node_potentials = d1**bx.nvar * nodes_pdf**bx.psi_q * PDF[i1]**(1-bx.psi_q)
        
        max_node = np.argmax(node_potentials)
        
        bx._logger(msg='explore!')
        return outside_nodes[max_node]
    
    def get_exploratory_radius(bx):
        # empirical rule to get desired behavior
        p_desired = np.exp(-np.sqrt(bx.q * bx.nsim))
        
        # get matematically clean radius of it
        r = bx.sball.get_r(p_desired)
        return r
        
        
ExplorationLog = namedtuple('ExplorationLog', (
    "nsim",
    "nvar",
    "nfacets",
    "r","R",
    "inner",
    "shell",
    "outer",
    "FORM_outside",
    "TwoFORM_outside",
    "orth_outside",
    "inside",
    "outside",
    "success_points",
    "failure_points",
    "success",
    "failure",
    "mix",
    "vertex_estimation",
    "weighted_vertex_estimation"
    ))        
        
        
class DumbExploration(_Exploration):
    holydays = 0
    
    #č hull musí být v G prostoru s auto_update=True
    def __init__(bx, sample_box, q=10):
        bx.sample_box = sample_box
        
        bx.direct_plan = quadpy.un.mysovskikh_1(bx.nvar).points
        
        bx.convex_hull = khull.Grick(bx.f_model, bx.direct_plan, nrandom=50)
        bx.q = q
        
        bx.ghull = Ghull(bx.convex_hull)
        bx.sball = sball.Sball(bx.nvar)
        
        bx.to_explore = 2
        bx.to_refine = 0 # never refine
        
        #č přece ponechame složku pro uživatelské odhady
        #č stm kód může semka něco ukladat
        bx.estimations = []
        
        #č vítejte nové uložiště odhadů.
        #č Odhady z stm kódu už ale nemají na tohle sahat
        if hasattr(bx, 'filename'):
            bx.box_estimations = Store.create(bx.filename + "_exp", ExplorationLog)
        else:
            bx.box_estimations = []
        
        bx.regen()
    
        
    def init_parameters(bx):
        """
        Returns dictionary of parameters the DiceBox was initialized with
        """
        return {'sample_box':bx.sample_box, 'hull':bx.hull, 'q':bx.q}
                 
    
    def __repr__(bx):
        return "%s(**%s)"%(bx.__class__.__name__,  repr(bx.init_parameters()))
        
    def __str__(bx):
        return "%s(%s)"%(bx.__class__.__name__,  str(bx.init_parameters()))
        
        
    
    
    def regen(bx):
        if bx.nsim > 0:
            bx.estimate_outside()
        bx._nsim = bx.nsim
                 
    
    
    
    
    def __len__(bx):
        return bx.sample_box.nsim
    
    def __getitem__(bx, slice):
        #č stačí vratit sample_box
        return bx.sample_box[slice]
    
    def __getattr__(dx, attr):
        if attr == 'dicebox':
            return dx
            
        #č branime sa rekurzii
        # defend against recursion
        #оӵ рекурсилы пезьдэт!
        if attr == 'sample_box':
            raise AttributeError
                
        #ё По всем вопросам обращайтесь 
        #ё на нашу горячую линию    
        else:
            return getattr(dx.sample_box, attr)
    
    
    # The DiceBox Observer 
    def _logger(self, *args, msg="", indent=0, **kwargs):
        if not kwargs:
            kwargs = "" #č ať se nám prázdné závorky nezobrazujou
        print(self.__class__.__name__ + ":", msg, *args, kwargs) 
        
        
        # inspired by Qt
    def connect(self, slot): self._logger = slot
    def disconnect(self): del(self._logger)
    
        
    # přidávání vzorků musí bejt explicitní!
    def add_sample(bx, input_sample):
        bx.sample_box.add_sample(input_sample)
        bx.increment(bx.sample_box[bx._nsim:])
        bx._nsim = bx.nsim
    
    
        
        
    def increment(bx, input_sample):
        bx.estimate_outside()
        
        bx.box_estimations.append(bx.get_pf_estimation())
    
                
                    
    
    
    
        
        
    
    def estimate_outside(bx):
        #č konečně mám pořádnou stejtful třídu
        #č pokud mám aspoň jednu tečku, tak už je mi šuma
        #č zda se konvexní obálka vytvořila, či nikoliv
        
        
        shell_estimation, global_stats = bx.ghull.get_shell_estimation()
        
        # shell_estimation  -22: inner, -3: shell, -11: outer
        bx.shell_estimation = shell_estimation
        
        #č kvůli názvu neleze do namedtuple
        global_stats['TwoFORM_outside'] = global_stats.pop('2FORM_outside')
        #č ndim se nelibí pandas
        global_stats['nvar'] = global_stats.pop('ndim')
        
        bx.global_stats = global_stats
        
        
        
        
        
    def get_pf_estimation(bx):
        global_stats = bx.global_stats
        orth_outside = global_stats['outside'] = global_stats['orth_outside']
        pf_inside = global_stats['inside'] = 1 - orth_outside
        
        #č takže celou skříňku prostě bereme jako simplex
        event, event_id, fr, wfr = sx.get_simplex_event(bx, weighting_space='G')
        # -1=outside, 0=success, 1=failure, 2=mix
        tri_estimation = {0:0, 1:0, 2:0}
        tri_estimation[event_id] = pf_inside
        
        vertex_estimation = pf_inside * fr
        weighted_vertex_estimation = pf_inside * wfr
        
        
        failsi = bx.failsi
        global_stats['success_points'] = len(failsi[~failsi])
        global_stats['failure_points'] = len(failsi[failsi])
        global_stats['success'] = tri_estimation[0]
        global_stats['failure'] = tri_estimation[1]
        global_stats['mix'] = tri_estimation[2]
        global_stats['vertex_estimation'] = vertex_estimation
        global_stats['weighted_vertex_estimation'] = weighted_vertex_estimation
        
        return ExplorationLog(**global_stats)
        



#!/usr/bin/env python
# coding: utf-8

import numpy as np
import scipy.special as sc
from scipy import stats

#######################################################
# s-balls  -- tools to calc probabilities and radii ###
#######################################################


def get_ps_ball(d,R):
    "returns probability of falling inside d-ball with radius R"
    #return np.sum(np.exp(-(rho**2)/2)*rho**(d-1) )* R/n
    return sc.gammainc(d/2, R**2/2)

def get_pf_ball(d,R):
    "returns probability of falling outside d-ball with radius R"
    return sc.gammaincc(d/2, R**2/2)

def get_Radius_pf(d,pf_ball):
    "returns radius of an d-ball with probability pf outside"
    rsqdiv2 = sc.gammainccinv(d/2, pf_ball)
    return np.sqrt(2*rsqdiv2) #radius

def get_Radius_ps(d,ps_ball):
    "returns radius of an d-ball with probability ps inside"
    rsqdiv2 = sc.gammaincinv(d/2, ps_ball)
    return np.sqrt(2*rsqdiv2) #radius





# implement class compatible to the old ones

# dispatcher
def Sball(nvar):
    if nvar == 2:
        return Sball_2D(nvar)
    else:
        return Sball_nD(nvar)

class Sball_nD:
    def __init__(self, nvar):
        self.nvar = nvar
        self.a = nvar/2
        
    def get_pf(self, r):
        "returns pf, i.e. complementary part of multidimensional Gaussian distribution"
        return sc.gammaincc(self.a, r**2/2)
        
    def get_ps(self, r):
        "returns probability of falling inside d-ball with radius R"
        return sc.gammainc(self.a, r**2/2)
    
    def get_r(self, desired_pf):
        "sball inversion. Returns radius of the s-ball with probability pf outside"
        rsqdiv2 = sc.gammainccinv(self.a, desired_pf)
        return np.sqrt(2*rsqdiv2) #radius
    
    
    def get_r_iteration(self, desired_pf):
        "Same as .get_r(), just keeps compatibility with previous versions"
        return self.get_r(desired_pf), desired_pf
    
    # make it, finally, scipy.stats -compatible
    sf = get_pf
    isf = get_r
    cdf = get_ps
    def ppf(self, q):
        return get_Radius_ps(self.nvar, q)

    
class Sball_2D(Sball_nD):
    def get_pf(self, r):
        "returns pf, i.e. complementary part of multidimensional Gaussian distribution"
        return np.exp(-r**2/2)
    
    def get_r(self, desired_pf):
        "sball inversion. Returns radius of the s-ball with probability pf outside"
        return np.sqrt(-2*np.log(desired_pf))
    
    
# calculation is as fast as Sball_nD
# but I'm not sure about precision
class Sball_4D(Sball_nD):
    def get_pf(self, r):
        "returns pf, i.e. complementary part of multidimensional Gaussian distribution"
        return (r**2/2+1)*np.exp(-r**2/2)








#1/ univariate funkce pro bounded Gauss
# left-right-bounded univariate Gaussian
class Radial:
    def __init__(self, nvar, r=0, R=np.inf):
        self.sball = Sball(nvar)
        self.set_bounds(r, R)
        
    def set_bounds(self, r=0, R=np.inf):
        #č kbyby se někomu nechtělo naťukat "np.inf"
        self.r = r # left bound
        self.R = R # rigth bound
        
        self.ps = self.sball.get_ps(r)
        self.pf = self.sball.get_pf(R)
        #č obsah pravděpodobnosti v mezikruži
        # well, probability falling to the shell
        self.p_shell = self.sball.get_pf(r) - self.pf
    
    #č jen pro formu. Kdo by to potřeboval?
    def _pdf(self, x): return stats.chi.pdf(x, self.sball.nvar) / self.p_shell
    def _cdf(self, x): return (self.sball.get_ps(x) - self.ps) / self.p_shell
    def _sf(self, x): return (self.sball.get_pf(x) - self.pf) / self.p_shell
    
    def pdf(self, x):
        return np.piecewise(x, [x<=self.r, x>=self.R], [0, 0, self._pdf])
        
    def cdf(self, x):
        return np.piecewise(x, [x<=self.r, x>=self.R], [0, 1, self._cdf])
        
    def sf(self, x): # 1 - cdf
        return np.piecewise(x, [x<=self.r, x>=self.R], [1, 0, self._sf])

    def ppf(self, q):
        return get_Radius_ps(self.sball.nvar, q*self.p_shell + self.ps)
        
    def isf(self, q): # inverse of .sf()
        return self.sball.get_r(q*self.p_shell + self.pf)
        
        

def get_random_directions(ns, ndim):
    # rand_dir: prepare ns random directions on a unit d-sphere
    rand_dir = np.random.randn(ns, ndim) #random directions
    
    
    lengths = np.sum(np.square(rand_dir), axis=1)
    lengths = np.sqrt(lengths, out=lengths) #lengths of each radius-vector
    
    # scale all radii-vectors to unit length
    # use [:,None] to get an transposed 2D array
    rand_dir = np.divide(rand_dir, lengths[:,None], out=rand_dir) 
    
    return rand_dir



#č nebyl to úplně ideální napad dědit od Radial
#č cdf, ppf, sf a isf metody nejsou pro Shell aplikovatelné!
#č Ještě jednou, bacha, davejte pozor, co vztahuje k 1D radiálnímu rozdělení,
#č co - k optimálnímu IS rozdělení proporcionálnímu nD Gaussu.
# We apologize for inconvenience
class Shell(Radial):
    """
    Optimal sampling density for Nv-ball (gaussian samples outside Nv-ball)
    with density proportional to Gaussian density
    """
    def rvs(self, size=1): # keyword size is scipy.stats-compatible
        "Generování­ vzorků (kandidátů a integračních bodů)"
        ns = size
        # rand_dir: prepare ns random directions on a unit d-sphere
        rand_dir = get_random_directions(ns, self.sball.nvar) #random directions
        
        # generate sampling probabilites
        p = np.linspace(1, 0, ns, endpoint=False) # probabilities for the radius
        
        # convert probabilitites into random radii
        # (distances from origin that are greater than r and less than R)
        r = self.isf(p) # actually, it is the same as CDF inverse
        
        #finally a random sample from the optimal IS density:
        sample_G = rand_dir*r[:,None]
        
        return sample_G

    #оӵ кулэ ӧвӧл #č multi nebudem použivat
    def _pdf_multi(self, x): return np.prod(stats.norm.pdf(x), axis=1)/self.p_shell
    
    def _pdf_norm(self, rx): 
        # stats.norm.pdf(0) == 0.3989422804014327
        c = 0.3989422804014327**(self.sball.nvar-1)
        return stats.norm.pdf(rx)*c/self.p_shell
    
    #č Vypocet hustot optimalni vzorkovaci veliciny (left-right-bounded)
    def pdf(self, x):
        """density of optimal IS variable h, 
        i.e. Gaussian variable with zero density inside d-ball"""
        rx = np.sum(x**2, axis=1)
        rx = np.sqrt(rx, out=rx) 
        return np.piecewise(rx, [rx<=self.r, rx>=self.R], [0, 0, self._pdf_norm])




#!/usr/bin/env python
# coding: utf-8

# One does not simply calculate the sample variance.
# https://github.com/numpy/numpy/issues/6231
# Only one citation from: 
# "Online calculations happen all the time (due to resource limitations), 
# and a np.welford function would be convenient so that not everyone 
# has to implement welford on its own over and over again."

# There are actually some code can be found in internet,
# but those I've seen supposed one-by-one data addition.

"""
This is implementation of generalized (for arbitrary sample sizes) 
Welford's updating (online) algorithm, given by Chan et al. in two papers:
"Algorithms for computing the sample variance: Analysis and recommendations"
https://doi.org/10.1080%2F00031305.1983.10483115
and
 "Updating Formulae and a Pairwise Algorithm for Computing Sample Variances."
http://i.stanford.edu/pub/cstr/reports/cs/tr/79/773/CS-TR-79-773.pdf
"""


import numpy as np



class Welford:
    def __init__(self):
        self.n = 0
        
        
    def add(self, data):
        n = len(data)
        T = np.sum(data)
        #č takhle je to numericky nejstabilnější
        S = np.sum(np.square(data - T/n))
        
        if self.n == 0:
            self.n = n
            self.T = T
            self.S = S
        else:
            m = self.n #č to, co bylo
            self.S += S + m /n /(m+n) * (n/m * self.T - T)**2
            self.n += n
            self.T += T
            
    #č Pro řídká data, speciálita pro IS
    #č uděláme explicitnou funkci s povinným sample size
    def add_sparse(self, data, n):
        """Method processes sparse data, assumes 
        only non-zero values of entire sample (of size n) need to be given"""
        sample_size = len(data)
        assert n >= sample_size
        
        T = np.sum(data)
        mean = T/n
        #č takhle je to numericky nejstabilnější
        S = np.sum(np.square(data - mean)) + (n - sample_size) * mean**2
        
        if self.n == 0:
            self.n = n
            self.T = T
            self.S = S
        else:
            m = self.n #č to, co bylo
            self.S += S + m /n /(m+n) * (n/m * self.T - T)**2
            self.n += n
            self.T += T
            
    @property
    def mean(self): 
        return self.T / self.n
    
    @property
    def var(self): 
        return self.S / self.n
        
    @property
    def s2(self): 
        return self.S / (self.n - 1)




#!/usr/bin/env python
# coding: utf-8

"""
tohle je mezivrstva mezi simplexovymi odhady v boxu a grafy v qt_plot
(Kdybych věřil, že je to jen pro qt_plot, tak ten modul nevytvařím)

#Podle kódu pyqtgraph a matplotlib už je jasný, že musím vracet 2 numpy vektory.
#Asi nebudu zahrnovat .zerosafe() - je to jen zvlaštnost pyqtgraph
#Není úplně jasně, zda mám dělat třídu (stateful), 
#nebo mám se omezit jen na funkci (stateless)
#Minule přišel jsem na potřebu stateful, 
#ale už si nepamatuji proč.

není aktuální, ale ty estimátory furt nemám promyšlený.
Podstata problému spočívá v tom, že současný DiceBox
ukladá odhady dvěma různejma způsoby, přičemž 
TRI odhady jsou ukladány jako slovníky (strukturováně).
Pro ty strukturovaná data Pandas (aj obecně tabulka)
není ideální volba. Na druhou stranu chceme jednoduše
všechno co máme nakreslit a vyexportovat do Excelu.
"""

import numpy as np
#č právě kvůli pandas davam tohle do separatního modulu
#č normálně kód WellMet pandas nevyžaduje
import pandas as pd 
        

def get_estimation_data(estimations, metric):
    metric_dict = dict()
    #č new-style: šecko leží dohromady a každý z toho
    #č bere co chce a jak chce
    #č ne že by to bylo nějak šetrný
    #č estimation je slovníkem
    for estimation in estimations:
        #č nsim musí mäť každej odhad
        #č pokud nemá - je třeba jej prostě opravit
        nsim = estimation['nsim']
        try: 
            metric_dict[nsim] = estimation[metric]
        except KeyError as e:
            pass #print(self.__class__.__name__ + ":", repr(e))
    
    #č nikdo neslibil, že budou v pořadí
    x = np.sort(tuple(metric_dict.keys()))
    y = np.array(tuple(metric_dict.values()))[np.argsort(tuple(metric_dict.keys()))]
    return x, y


def proxy(dice_box, nsim):
    proxy = dice_box.proxy
    index = np.array(nsim)-1
    #č indexy musíme o jedničku změnšit
    #č výsledek nikoliv. Takže v cajku.
    return np.cumsum(~proxy.astype(bool))[index]


#č vysledek dlouhého přemyšlení, co a jak mám udělat.
#č nic lepšího nenapadá: 
#č pandas není vhodný pro strukturována data
#č (možná aj strukturování není úplně na místě), 
#č ale exportovat do Excelčíku se mi taky chce
def get_tri_data_frame(dice_box, sources=['box', 'user'], apply_proxy=False):
    
    #č přeneseme rozhodování do volajícího kódu
#    #č nejdřív proxy. None znamená podle přitomosti
#    if apply_proxy is None:
#        if hasattr(dice_box, 'proxy'):
#            apply_proxy = True
#        else:
#            apply_proxy = False
    
    #č teď zdroje (zjednodušeně). Zatím netřeba nic komplikovat
    #č po těch pomocných funkcích budu chtit df s nastaveným index=nsim
    #č a sloupcem nsim (ten je spíš pro nás, lidé). proxy nemají řešit
    
    #č je tu fakt velkej bordel s těmi odhadama
    #č to jistě budu muset překopávat
    tri_box_estimator = 'TRI_overall_estimations'
    if ('box' in sources) and (tri_box_estimator in dice_box.guessbox.estimations):
        df_box = get_tri_box_df(dice_box, tri_box_estimator)
    else: #č pak nakrmíme pd.concat'u
        df_box = None
    
    if 'user' in sources:
        df_user = get_tri_user_df(dice_box)
    else: #č pak nakrmíme pd.concat'u
        df_user = None
    
    #č předbežně:
    #č nejdřív concat (spojime zdroje dohromady)
    df = pd.concat((df_box, df_user), sort=False)
    
    #č pak deduplicate 
    #č (aby se převzaly odhady se stejným nsim z posledního zdroje)
    df = df.groupby(level=0).last()
    
    #č dale vytřídíme odhady dle nsim
    df.sort_values('nsim', inplace=True)
    
    #č pokud použiváme proxy, tak je vložit, vyhodit ďupy, nahradit index
    if apply_proxy:
        #č teoreticky index a nsim musejí bejt stejné
        nsim = df.index
        nsim_proxy = proxy(dice_box, nsim)
        #č oboje nsim a nsim_proxy jsou pro lide, aby zahrivalo srdce
        #č předpokladá se, že volající kód bude použivat index
        df.insert(loc=0, column='nsim (proxy)', value=nsim_proxy)
        #č nahradit index, pak vyhodit ďupy
        df.index = nsim_proxy
        df = df.groupby(level=0).last()
        
    #č když ne - tak jenom nahradíme index a vypadneme otsuď 
    #č Alexi, co? vždyť ten index už nemusíme řešit, ten musí bejt v pořádku!
    
    #č vemte ten svůj pitomej rám
    #č a na shledanou!
    return df
    
    
    
    

def get_tri_box_df(dice_box, tri_estimation_name='TRI_overall_estimations'):
    #č chyby nechť chytá volající kód!
    data = dice_box.guessbox.estimations[tri_estimation_name]
    nsim, tri_data = data
    # it can be effectively done with pandas
    df = pd.DataFrame(tri_data, index=nsim)
    # -1 = 'outside', 0=success, 1=failure, 2=mix
    df.rename(columns={-1:'outside', 0:'success', 1:'failure', 2:'mix'}, inplace=True)
    df.insert(loc=0, column='nsim', value=nsim)
    
    
    if 'vertex_estimation' in dice_box.guessbox.estimations:
        data = dice_box.guessbox.estimations['vertex_estimation']
        nsim, y = data
        #č zatím nevidím důvod komplikovat
        #č ale je tu předpoklad konzistenci skříňky
        df['vertex_estimation'] = y 
    
    if 'weighted_vertex_estimation' in dice_box.guessbox.estimations:
        data = dice_box.guessbox.estimations['weighted_vertex_estimation']
        nsim, y = data
        #č zatím nevidím důvod komplikovat
        #č ale je tu předpoklad konzistenci skříňky
        df['weighted_vertex_estimation'] = y 
        
    return df
            




def get_tri_user_df(dice_box):
    metrics = ['TRI_estimation', \
                 'vertex_estimation', 'weighted_vertex_estimation']
                            
    metadict = dict()
    for metric in metrics:
        metadict[metric] = dict()
    
    #č new-style: šecko leží dohromady a každý si z toho
    #č bere co chce a jak chce
    #č ne že by to bylo nějak šetrný
    #č estimation je slovníkem
    for estimation in dice_box.estimations:
        # nsim musí mäť každej odhad
        # pokud nemá - je třeba jej prostě opravit
        nsim = estimation['nsim']
        for metric in metrics:
            #č tuším __contains__ je lehčí jak exception
            if metric in estimation:
                metadict[metric][nsim] = estimation[metric]
    
    #č vytahneme 'TRI_estimation' z metrik aj ze slovníku
    tri_estimation = metadict.pop(metrics.pop(0)) # == 
    #č zda se, že zde nic nehodí chybu, aj kdyby žádné odhady nebyly
    nsim = tuple(tri_estimation.keys())
    # it can be effectively done with pandas
    df = pd.DataFrame(tuple(tri_estimation.values()), index=nsim)
    # -1 = 'outside', 0=success, 1=failure, 2=mix
    df.rename(columns={-1:'outside', 0:'success', 1:'failure', 2:'mix'}, inplace=True)
    df.insert(loc=0, column='nsim', value=nsim)
    #df.sort_values('nsim', inplace=True) #č zbytečně
    
    #č musely tam zůstat jenom 'vertex_estimation' 
    #č a 'weighted_vertex_estimation'
    for metric in metrics:
        s = pd.Series(metadict[metric])
        df[metric] = s
    
    return df

#č sice nazev obsahuje DataFrame, dědiť my jeho, конечно же, nebudeme
#č pořadně nerozumím, proč tu třídu dělám. Valčím s komplicitou
#č zatím jsem rozhod, že je zbytečně dělat wrap .pf_exact'u a .pf_exact_method'u
#č patří navíc nahé skřiňce
#class BoxEstimationsDataFrame:
#    def __init__(self, dice_box):
#        self.dice_box = dice_box
#        
#        
#            
#    
#    #č aspoň něco
#    def __contains__(self, estimation):
#          return True 
#        
#            
#
#    def _pens_data_update(self):
#        df = self.df
#        nsim = df.nsim.to_numpy()
#        if self.proxy_chk.isChecked():
#            x = self.proxy(nsim)
#            df.insert(loc=0, column='nsim (proxy)', value=x)
#        else:
#            x = nsim
#        # (in case of LogPlot) fallback values also used
#        success_values = df.failure+df.mix+df.out
#        outmix_values = df.failure+df.mix
#        failure_fallback = np.where(outmix_values > 0, outmix_values, success_values)
#        self.pen_f.setData(*self.zerosafe(x, df.failure, failure_fallback))
#        self.pen_outmix.setData(*self.zerosafe(x, outmix_values, success_values))
#        self.pen_success.setData(*self.zerosafe(x, success_values))
#    
#    
#    def redraw(self):
#        xmin = np.inf
#        xmax = -np.inf
#        tri_estimation = dict()
#        try: # тут всё что угодно может пойти не так
#            # kruci, ještě navic i generovať pokažde znovu...
#        
#            # new-style: šecko leží dohromady a každý si z toho
#            # bere co chce a jak chce
#            # ne že by to bylo nějak šetrný
#            # estimation je slovníkem
#            for estimation in self.dice_box.estimations:
#                # nsim musí mäť každej odhad
#                # pokud nemá - je třeba jej prostě opravit
#                nsim = estimation['nsim']
#                try: 
#                    tri_estimation[nsim] = estimation['TRI_estimation']
#                    if nsim > xmax:
#                        xmax = nsim
#                    if nsim < xmin:
#                        xmin = nsim
#                        
#                except KeyError as e:
#                    pass #print(self.__class__.__name__ + ":", repr(e))
#            
#            #č neotravuj uživatele chybovejma hlaškama
#            if tri_estimation:
#                # it can be effectively done with pandas
#                self.df = df = pd.DataFrame(tuple(tri_estimation.values()))
#                # -1 = 'out', 0=success, 1=failure, 2=mix
#                df.rename(columns={-1:'out', 0:'success', 1:'failure', 2:'mix'}, inplace=True)
#                df.insert(loc=0, column='nsim', value=tuple(tri_estimation.keys()), allow_duplicates=False)
#                df.sort_values('nsim', inplace=True)
#
#                self._pens_data_update()
#            
#                nsim, y = get_estimation_data(self.dice_box.estimations, 'vertex_estimation')
#                df['vertex_estimation'] = y #č spolehám na konzistenci odhadů (ne úplně)
#                self.pen_vertex.setData(*self.zerosafe(self.proxy(nsim), y))
#                
#                nsim, y = get_estimation_data(self.dice_box.estimations, 'weighted_vertex_estimation')
#                df['weighted_vertex_estimation'] = y #č spolehám na konzistenci odhadů (ne úplně)
#                self.pen_weighted_vertex.setData(*self.zerosafe(self.proxy(nsim), y))
#        
#            
#        except BaseException as e:
#            print(self.__class__.__name__ + ":", repr(e))
#        



#!/usr/bin/env python
# coding: utf-8


from . import qt_box_functions as gui


wt = gui.select_test_case()

#č funkce read_box() buď vytvoří Reader object
#č kterým nahradí wt.sample_box,
#č nebo - je-li uživatel odmitné zvolit nějaký soubor
#č ponechá původní samplebox
gui.read_box(wt)

gui.setup_dicebox(wt)

gui.show_box(wt)


